from typing import Any, Tuple, List, Dict

from einops import rearrange, repeat, reduce
from torch import nn, Tensor
from torch.nn import functional
from torchvision.ops import masks_to_boxes
import torch

from sihl.layers import ConvNormAct
from sihl.utils import coordinate_grid, interpolate

from .object_detection import ObjectDetection


class InstanceSegmentation(ObjectDetection):
    """Instance segmentation is like object detection except instances are associated
    with a binary mask instead of a bounding box.

    Refs:
        1. [CondInst](https://arxiv.org/abs/2102.03026)
    """

    def __init__(
        self,
        in_channels: List[int],
        num_classes: int,
        bottom_level: int = 3,
        top_level: int = 7,
        num_channels: int = 256,
        num_layers: int = 4,
        max_instances: int = 100,
        t_min: float = 0.2,
        t_max: float = 0.6,
        topk: int = 7,
        soft_label_decay_steps: int = 1,
        mask_top_level: int = 5,
        mask_num_channels: int = 8,
        mask_out_channels: int = 1,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_classes (int): Number of possible object categories.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            max_instances (int, optional): Maximum number of instances to predict in a sample. Defaults to 100.
            t_min (float, optional): Lower bound of O2F parameter t. Defaults to 0.2.
            t_max (float, optional): Upper bound of O2F parameter t. Defaults to 0.6.
            topk (int, optional): How many anchors to match to each ground truth object when copmuting the loss. Defaults to 7.
            soft_label_decay_steps (int, optional): How many training steps to perform before the one-to-few matching becomes one-to-one. Defaults to 1.
            mask_top_level (int, optional): Top level of inputs masks are computed from. Defaults to 5.
            mask_num_channels (int, optional): Number of convolutional channels in the mask prediction branch.
            mask_out_channels (int, optional): Number of output channels in the mask prediction branch.
        """
        assert top_level >= mask_top_level >= bottom_level
        super().__init__(
            in_channels=in_channels,
            num_classes=num_classes,
            bottom_level=bottom_level,
            top_level=top_level,
            num_channels=num_channels,
            num_layers=num_layers,
            max_instances=max_instances,
            t_min=t_min,
            t_max=t_max,
            topk=topk,
            soft_label_decay_steps=soft_label_decay_steps,
        )

        self.mask_top_level = mask_top_level
        self.mask_num_channels = mask_num_channels
        self.mask_out_channels = mask_out_channels
        c, o = mask_num_channels, mask_out_channels
        controller_out_channels = (c + 2) * c + c + c * c + c + c * o + o
        self.controller_head = nn.Conv2d(num_channels, controller_out_channels, 1)
        bl = bottom_level
        self.bottom_module = ConvNormAct(in_channels[bl], self.mask_num_channels)
        self.output_shapes = {
            "num_instances": ("batch_size",),
            "scores": ("batch_size", max_instances),
            "classes": ("batch_size", max_instances),
            "masks": ("batch_size", max_instances, f"height/{2**bl}", f"width/{2**bl}"),
        }

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        num_instances, scores, classes, logits = self.forward_logits(inputs)
        masks = interpolate(logits.sigmoid(), size=inputs[0].shape[2:])
        return num_instances, scores, classes, masks

    def forward_logits(
        self, inputs: List[Tensor]
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        scores, reg_features = self.get_features(inputs)
        scores, pred_classes = scores.max(dim=2)
        # keep only the top `max_instances`
        batch_size = scores.shape[0]
        batches = repeat(torch.arange(batch_size), f"b -> b {self.max_instances}")
        instance_idxs = scores.topk(self.max_instances).indices
        scores = scores[batches, instance_idxs]
        pred_classes = pred_classes[batches, instance_idxs]
        num_instances = reduce(scores > self.threshold, "b i -> b", "sum")

        # compute segmentation masks for the `max_instances` top scoring anchors
        controllers = self.get_controllers(reg_features)
        bottom_features = self.get_bottom_features(inputs)
        normalized_anchors = self.get_anchors(inputs, normalized=True)
        logits = torch.cat(
            [
                self.get_masks(
                    params=controllers[batch_idx, instance_idxs[batch_idx]],
                    anchors=normalized_anchors[instance_idxs[batch_idx]],
                    bottom_features=bottom_features[batch_idx],
                    num_masks=self.max_instances,
                )
                for batch_idx in range(batch_size)
            ]
        )
        return num_instances, scores, pred_classes, logits

    def get_controllers(self, features: List[Tensor]) -> Tensor:
        controllers = [self.controller_head(_) for _ in features]
        return torch.cat(
            [rearrange(_, "n c h w -> n (h w) c") for _ in controllers], dim=1
        )

    def get_bottom_features(self, inputs: List[Tensor]) -> Tensor:
        assert len(inputs) > self.top_level
        x0 = inputs[self.bottom_level]
        batch_size, _, height, width = x0.shape
        summed_features = torch.stack(
            [x0]
            + [
                interpolate(inputs[level], size=(height, width))
                for level in range(self.bottom_level + 1, self.mask_top_level + 1)
            ]
        ).sum(dim=0)
        bottom_feats = self.bottom_module(summed_features)
        # Absolute coordinates will later be transformed into relative ones
        abs_coords = coordinate_grid(height, width, y_max=1, x_max=1)
        abs_coords = repeat(abs_coords, "c h w -> b c h w", b=batch_size, c=2)
        return torch.cat([bottom_feats, abs_coords.to(bottom_feats)], dim=1)

    def get_masks(
        self, params: Tensor, anchors: Tensor, bottom_features: Tensor, num_masks: int
    ) -> Tensor:
        c, m_c = self.mask_num_channels, num_masks * self.mask_num_channels
        o, m_o = self.mask_out_channels, num_masks * self.mask_out_channels

        w1 = params[:, (s := slice(0, (c + 2) * c))].reshape(m_c, c + 2, 1, 1)
        b1 = params[:, (s := slice(s.stop, s.stop + c))].reshape(m_c)
        w2 = params[:, (s := slice(s.stop, s.stop + c**2))].reshape(m_c, c, 1, 1)
        b2 = params[:, (s := slice(s.stop, s.stop + c))].reshape(m_c)
        w3 = params[:, (s := slice(s.stop, s.stop + c * o))].reshape(m_o, c, 1, 1)
        b3 = params[:, s.stop :].reshape(m_o)

        # Convert abs coords to rel coords by subtracting padded anchors
        padded_anchors = functional.pad(anchors, (c, 0, 0, 0)).reshape(-1, 1, 1)
        x = repeat(bottom_features, "c h w -> (n c) h w", n=num_masks) - padded_anchors
        x = functional.conv2d(x.unsqueeze(0), w1, b1, groups=num_masks).relu()
        x = functional.conv2d(x, w2, b2, groups=num_masks).relu()
        return functional.conv2d(x, w3, b3, groups=num_masks)

    def training_step(
        self,
        inputs: List[Tensor],
        classes: List[Tensor],
        masks: List[Tensor],
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, Any]]:
        boxes = [masks_to_boxes(sample_masks) for sample_masks in masks]
        scores, reg_features = self.get_features(inputs)
        controllers = self.get_controllers(reg_features)
        pred_boxes = self.get_boxes(reg_features)
        bottom_feats = self.get_bottom_features(inputs)
        anchors = self.get_anchors(inputs, normalized=False)
        class_target, box_target, assignment = self.get_targets(
            anchors, scores, pred_boxes, classes, boxes, is_validating
        )
        class_loss, box_loss = self.get_losses(
            scores, pred_boxes, class_target, box_target
        )
        mask_losses = []  # dice loss
        batch_size = scores.shape[0]
        normed_anchors = self.get_anchors(inputs, normalized=True)
        pos_mask = reduce(class_target, "b i s -> b i", "max") == 1.0
        for batch_idx in range(batch_size):
            pos_idxs = pos_mask[batch_idx]
            if masks[batch_idx].numel() == 0 or not torch.any(pos_idxs):
                continue
            pred_masks = self.get_masks(
                params=controllers[batch_idx, pos_idxs],
                anchors=normed_anchors[pos_idxs],
                bottom_features=bottom_feats[batch_idx],
                num_masks=pos_idxs.sum(),
            ).sigmoid()
            target_masks = masks[batch_idx][assignment[batch_idx, pos_idxs]]
            mask_size = (target_masks.shape[1], target_masks.shape[2])  # FIXME ?
            pred_masks = interpolate(pred_masks, size=mask_size).squeeze(0)
            numerator = reduce(pred_masks * target_masks, "c h w -> c", "sum")
            denominator = reduce(pred_masks**2 + target_masks**2, "c h w -> c", "sum")
            mask_losses.append((1 - 2 * numerator / denominator).nan_to_num(0).mean())
        mask_loss = torch.stack(mask_losses).mean()
        return class_loss + box_loss + mask_loss, {
            "class_loss": class_loss,
            "box_loss": box_loss,
            "mask_loss": mask_loss,
        }

    def validation_step(
        self, inputs: List[Tensor], classes: List[Tensor], masks: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, Any]]:
        loss, metrics = self.training_step(inputs, classes, masks, is_validating=True)
        self.loss_computer.to(loss.device).update(loss)
        boxes = [masks_to_boxes(sample_masks) for sample_masks in masks]
        num_instances, scores, pred_classes, pred_masks = self.forward(inputs)
        map_computer_preds = []
        for batch_idx, sample_masks in enumerate(pred_masks):
            non_empty_idxs = torch.any(sample_masks > 0.5, dim=(1, 2))
            sample_boxes = masks_to_boxes(sample_masks[non_empty_idxs] > 0.5)
            map_computer_preds.append(
                {
                    "scores": scores[batch_idx][non_empty_idxs],
                    "labels": pred_classes[batch_idx][non_empty_idxs],
                    "boxes": sample_boxes,
                }
            )
        self.map_computer.to(scores.device).update(
            map_computer_preds,
            [{"labels": c.to(torch.int64), "boxes": b} for c, b in zip(classes, boxes)],
        )
        return loss, metrics



import torch
from torch import nn, Tensor


class SpatialAttention(nn.Module):
    """https://arxiv.org/abs/1807.06521"""

    def __init__(self, kernel_size: int = 7) -> None:
        super().__init__()
        self.conv = nn.Conv2d(2, 1, kernel_size, padding=kernel_size // 2)

    def forward(self, x: Tensor) -> Tensor:
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out = torch.max(x, dim=1, keepdim=True).values
        return self.conv(torch.cat([avg_out, max_out], dim=1)).sigmoid()


class ChannelAttention(nn.Module):
    """https://arxiv.org/abs/1807.06521"""

    def __init__(self, in_channels: int, ratio: int = 16) -> None:
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, in_channels // ratio, kernel_size=1),
            nn.ReLU(),
            nn.Conv2d(in_channels // ratio, in_channels, kernel_size=1),
        )

    def forward(self, x: Tensor) -> Tensor:
        avg_out = self.conv(self.avg_pool(x))
        max_out = self.conv(self.max_pool(x))
        return (avg_out + max_out).sigmoid()


class CBAM(nn.Module):
    """https://arxiv.org/abs/1807.06521"""

    def __init__(
        self,
        in_channels: int,
        kernel_size: int = 7,
        ratio: int = 16,
        applied: bool = False,
    ) -> None:
        super().__init__()
        self.applied = applied
        self.channel_attention = ChannelAttention(in_channels, kernel_size)
        self.spatial_attention = SpatialAttention(kernel_size)

    def forward(self, x: Tensor) -> Tensor:
        x = self.channel_attention(x) * x
        attention = self.spatial_attention(x) * x
        return attention * x if self.applied else attention


class CrossCBAM(nn.Module):
    """https://arxiv.org/abs/2306.02306"""

    def __init__(self, in_channels: int, kernel_size: int = 7, ratio: int = 16) -> None:
        super().__init__()
        self.channel_attention_high = ChannelAttention(in_channels, ratio)
        self.spatial_attention_high = SpatialAttention(kernel_size)
        self.channel_attention_low = ChannelAttention(in_channels, ratio)
        self.spatial_attention_low = SpatialAttention(kernel_size)

    def forward(self, low: Tensor, high: Tensor) -> Tensor:
        mid_low = self.channel_attention_low(low) * high
        mid_high = self.channel_attention_high(high) * low
        out_low = self.spatial_attention_low(mid_low) * mid_high
        out_high = self.spatial_attention_high(mid_high) * mid_low
        return out_low + out_high



from typing import List

from matplotlib import pyplot as plt
import numpy as np
import torch

from sihl.heads import MultilabelClassification

from .common import get_images, plot_to_numpy


@get_images.register(MultilabelClassification)
def _(head, config, input, target, features) -> List[np.ndarray]:
    categories = config["categories"] if "categories" in config else None
    prediction = head(features)
    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch, image in enumerate(images):
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.axis("off")
        axes[0].title.set_text("Input")
        axes[0].imshow(image)

        def set_text(ax, text):
            ax.text(
                0.5,
                0.5,
                text,
                transform=ax.transAxes,
                ha="center",
                va="center",
                fontsize="large",
                bbox=dict(edgecolor="black", facecolor="white", boxstyle="round"),
            )

        axes[1].title.set_text("Target")
        axes[1].imshow(np.full_like(image, fill_value=255))
        if target is not None:
            labels = target[batch].to(torch.bool)
            patch_text = ""
            for idx, is_true in enumerate(labels):
                if is_true:
                    patch_text += str(idx) if categories is None else categories[idx]
                    patch_text += "\n"
            set_text(axes[1], patch_text.strip())
        axes[2].title.set_text("Prediction")
        axes[2].imshow(np.full_like(image, fill_value=255))
        if prediction is not None:
            scores = prediction[0][batch].cpu().numpy()
            labels = prediction[1][batch].cpu().numpy()
            patch_text = ""
            for score, label in zip(scores, labels):
                if score < 0.5:
                    continue
                patch_text += str(label) if categories is None else categories[label]
                patch_text += "\n"
            set_text(axes[2], patch_text.strip())
        # fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from pathlib import Path
from typing import Tuple, List, Literal, Dict
import csv
import logging

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision

from sihl import SihlModel, SihlLightningModule
from sihl.heads import AutoregressiveTextRecognition
from sihl.torchvision_backbone import TorchvisionBackbone
from sihl.utils import random_pad


def collate_fn(batch: List[Tuple[Tensor, Dict]]):
    batch = (
        torch.stack([sample[0] for sample in batch]),
        [[sample[1] for sample in batch]],  # (num_heads, num_samples)
    )
    return batch


class CyrillicDataset(torch.utils.data.Dataset[Tuple[Tensor, bool]]):
    max_length = 40
    tokens = [
        " ",
        "!",
        '"',
        "%",
        "'",
        "(",
        ")",
        "+",
        ",",
        "-",
        ".",
        "/",
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        ":",
        ";",
        "=",
        "?",
        "R",
        "[",
        "]",
        "b",
        "c",
        "e",
        "h",
        "i",
        "o",
        "p",
        "r",
        "s",
        "t",
        "u",
        "x",
        "y",
        "«",
        "»",
        "А",
        "Б",
        "В",
        "Г",
        "Д",
        "Е",
        "Ж",
        "З",
        "И",
        "Й",
        "К",
        "Л",
        "М",
        "Н",
        "О",
        "П",
        "Р",
        "С",
        "Т",
        "У",
        "Ф",
        "Х",
        "Ц",
        "Ч",
        "Ш",
        "Щ",
        "Э",
        "Ю",
        "Я",
        "а",
        "б",
        "в",
        "г",
        "д",
        "е",
        "ж",
        "з",
        "и",
        "й",
        "к",
        "л",
        "м",
        "н",
        "о",
        "п",
        "р",
        "с",
        "т",
        "у",
        "ф",
        "х",
        "ц",
        "ч",
        "ш",
        "щ",
        "ъ",
        "ы",
        "ь",
        "э",
        "ю",
        "я",
        "ё",
        "№",
    ]

    def __init__(self, data_dir: Path, split: Literal["test", "train"]) -> None:
        self.images: List[Tuple[Path, str]] = []
        images_dir = data_dir / split
        with open(data_dir / f"{split}.tsv") as f:
            reader = csv.reader(f, delimiter="\t")
            for line in reader:
                image_path = images_dir / line[0]
                assert image_path.exists()
                self.images.append((image_path, line[1]))

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, target = self.images[idx]
        image = torchvision.io.read_image(
            str(image_path.resolve()), torchvision.io.ImageReadMode.RGB
        )
        image = random_pad(image, target_size=(64, 256), fill=0).to(torch.float) / 255
        target = torch.tensor([self.tokens.index(char) for char in target])
        return image, target

    def __len__(self) -> int:
        return len(self.images)


class CyrillicDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int = 32) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "Cyrillic"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "constantinwerner/cyrillic-handwriting-dataset",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "fit") -> None:
        self.trainset = CyrillicDataset(self.data_dir, split="train")
        self.valset = CyrillicDataset(self.data_dir, split="test")

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, bool]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            drop_last=True,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, bool]]:
        return DataLoader(
            self.valset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            drop_last=True,
            collate_fn=collate_fn,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="autoregressive_text_recognition"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_epochs=50,
        accelerator="gpu",
        gradient_clip_val=1,
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = AutoregressiveTextRecognition(
            in_channels=backbone.out_channels,
            max_sequence_length=CyrillicDataset.max_length,
            num_tokens=len(CyrillicDataset.tokens),
            level=5,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.AdamW,
            optimizer_kwargs={"lr": 1e-4, "weight_decay": 1e-2},
            data_config={"tokens": CyrillicDataset.tokens},
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CyrillicDataModule(batch_size=64))



# ruff: noqa: F401
from typing import Protocol, Any, List, Dict, Tuple, Union

from torch import Tensor

from sihl.heads.anomaly_detection import AnomalyDetection
from sihl.heads.autoencoding import Autoencoding
from sihl.heads.autoregressive_text_recognition import AutoregressiveTextRecognition
from sihl.heads.depth_estimation import DepthEstimation
from sihl.heads.instance_segmentation import InstanceSegmentation
from sihl.heads.keypoint_detection import KeypointDetection
from sihl.heads.metric_learning import MetricLearning
from sihl.heads.multiclass_classification import MulticlassClassification
from sihl.heads.multilabel_classification import MultilabelClassification
from sihl.heads.object_detection import ObjectDetection
from sihl.heads.panoptic_segmentation import PanopticSegmentation
from sihl.heads.quadrilateral_detection import QuadrilateralDetection
from sihl.heads.regression import Regression
from sihl.heads.scene_text_recognition import SceneTextRecognition
from sihl.heads.semantic_segmentation import SemanticSegmentation
from sihl.heads.view_invariance_learning import ViewInvarianceLearning


TensorShape = Tuple[Union[str, int], ...]
Loss = Tensor
Metrics = Dict[str, float]


class Head(Protocol):
    output_shapes: Dict[str, TensorShape]

    def forward(self, inputs: List[Tensor]) -> Any:
        """Performs an inference pass.
        This function is traced and converted when exporting to ONNX.

        Args:
            inputs (List[Tensor]): Input tensors by feature level

        Returns:
            Any: Depends on the head
        """
        pass

    def training_step(self, inputs: List[Tensor], *args: Any) -> Tuple[Loss, Metrics]:
        pass

    def on_validation_start(self) -> None:
        pass

    def validation_step(self, inputs: List[Tensor], *args: Any) -> Tuple[Loss, Metrics]:
        pass

    def on_validation_end(self) -> Metrics:
        pass



from pathlib import Path
from typing import Tuple, List
from random import shuffle
import logging
import sys

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

sys.path.append("../src")


from sihl.heads import Regression
from sihl import SihlModel, SihlLightningModule
from sihl.torchvision_backbone import TorchvisionBackbone


class AgePredictionDataset(torch.utils.data.Dataset[Tuple[Tensor, Tensor]]):
    def __init__(self, data_dir: Path, train: bool = False) -> None:
        self.train = train
        self.data_dir = data_dir
        self.images_and_targets: List[Tuple[Path, float]] = []
        for subdir in data_dir.iterdir():
            age = float(subdir.name)
            for imgf in subdir.glob("*.jpg"):
                self.images_and_targets.append((imgf, age))
        shuffle(self.images_and_targets)
        if self.train:
            self.preprocess = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(p=0.5),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.preprocess = transforms.ToDtype(torch.float32, scale=True)

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, target = self.images_and_targets[idx]
        image = torchvision.io.read_image(
            str(image_path.resolve()), mode=torchvision.io.ImageReadMode.RGB
        )
        image = self.preprocess(image)
        return image, torch.tensor(target)

    def __len__(self) -> int:
        return len(self.images_and_targets)


class AgePredictionDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "AgePrediction"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "mariafrenti/age-prediction",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = AgePredictionDataset(
            self.data_dir / "age_prediction_up" / "age_prediction" / "train", train=True
        )
        self.valset = AgePredictionDataset(
            self.data_dir / "age_prediction_up" / "age_prediction" / "test", train=False
        )

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset, batch_size=self.batch_size, shuffle=False, num_workers=4
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="regression"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_epochs=10,
        accelerator="gpu",
        val_check_interval=0.5,
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        precision="16-mixed",
    )

    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = Regression(
            in_channels=backbone.out_channels, lower_bound=0, upper_bound=100
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.AdamW,
            optimizer_kwargs={"lr": 1e-4, "weight_decay": 1e-4},
        )

    log.debug(
        summary(
            model.model,
            input_data=model.model.backbone.dummy_input,
            row_settings=("var_names",),
            col_names=("output_size", "num_params", "mult_adds", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=AgePredictionDataModule(batch_size=64))



from pathlib import Path
from typing import List

from torch import Tensor
import numpy as np
import onnx
import onnxruntime
import pytest
import torch

from sihl.heads import MultilabelClassification


BATCH_SIZE, NUM_CHANNELS, HEIGHT, WIDTH = 4, 512, 10, 10
NUM_LABELS, LEVEL = 100, 5
ONNX_VERSION, ONNX_FILE_NAME = 18, "multilabel_classification.onnx"


@pytest.fixture()
def model() -> MultilabelClassification:
    return MultilabelClassification(
        in_channels=(3,) + (NUM_CHANNELS,) * LEVEL, num_labels=NUM_LABELS, level=LEVEL
    )


@pytest.fixture()
def backbone_output() -> List[Tensor]:
    return [torch.randn((BATCH_SIZE, 3, HEIGHT, WIDTH))] + [
        torch.randn((BATCH_SIZE, NUM_CHANNELS, HEIGHT, WIDTH)) for _ in range(LEVEL)
    ]


@pytest.fixture()
def targets() -> Tensor:
    return torch.randint(0, 2, (BATCH_SIZE, NUM_LABELS)).to(torch.float)


def test_forward(
    model: MultilabelClassification, backbone_output: List[Tensor]
) -> None:
    scores, labels = model(backbone_output)
    assert tuple(scores.shape) == (BATCH_SIZE, NUM_LABELS)
    assert tuple(labels.shape) == (BATCH_SIZE, NUM_LABELS)


def test_training_step(
    model: MultilabelClassification, backbone_output: List[Tensor], targets: Tensor
) -> None:
    loss, _ = model.training_step(backbone_output, targets)
    assert loss.item()


def test_validation_step(
    model: MultilabelClassification, backbone_output: List[Tensor], targets: Tensor
) -> None:
    model.on_validation_start()
    loss, _ = model.validation_step(backbone_output, targets)
    assert loss.item()
    metrics = model.on_validation_end()
    assert metrics


@pytest.fixture()
def onnx_model(model: MultilabelClassification, backbone_output: List[Tensor]) -> None:
    torch.onnx.export(
        model.to(torch.float32),
        args=backbone_output,
        f=ONNX_FILE_NAME,
        opset_version=ONNX_VERSION,
        input_names=[f"input_level_{idx}" for idx in range(len(backbone_output))],
        output_names=[f"head0/{name}" for name, shape in model.output_shapes.items()],
        dynamic_axes=dict(
            {
                f"input_level_{lvl}": {
                    0: "batch_size",
                    2: f"height/{2**lvl}",
                    3: f"width/{2**lvl}",
                }
                for lvl in range(len(backbone_output))
            },
            **{
                f"head0/{name}": {
                    shape_idx: str(shape_value)
                    for shape_idx, shape_value in enumerate(shape)
                }
                for name, shape in model.output_shapes.items()
            },
        ),
        external_data=False,
        verify=True,
        # dynamo=True,
        # report=True,
    )
    onnx_model = onnx.load(ONNX_FILE_NAME)
    Path(ONNX_FILE_NAME).unlink()
    return onnx_model


def test_onnx_inference(
    onnx_model, model: MultilabelClassification, backbone_output: List[Tensor]
) -> None:
    model.eval()
    onnx_session = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    onnx_input = {
        f"input_level_{idx}": _.numpy()
        for idx, _ in enumerate(backbone_output)
        if f"input_level_{idx}" in [node.name for node in onnx_model.graph.input]
    }
    # just check that 99% of values are equal.
    pytorch_output = [_.detach().numpy() for _ in model(backbone_output)]
    onnx_output = onnx_session.run(None, onnx_input)
    for i in range(len(pytorch_output)):
        assert (
            np.sum(np.abs(pytorch_output[i] - onnx_output[i]) > 1) / onnx_output[i].size
            < 0.01
        )



from torch import nn, Tensor
from torch.nn import functional
import numpy as np
import torch


class BlurPool2d(nn.Module):
    """https://arxiv.org/abs/1904.11486"""

    def __init__(self, in_channels: int, kernel_size: int = 3, stride: int = 1) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = [conv_padding(kernel_size, stride, dilation=1)] * 4
        coeffs = torch.tensor(
            (np.poly1d((0.5, 0.5)) ** (kernel_size - 1)).coeffs.astype(np.float32)
        )
        kernel = (coeffs[:, None] * coeffs[None, :])[None, None, :, :]
        self.register_buffer("kernel", kernel.repeat(in_channels, 1, 1, 1))

    def forward(self, x: Tensor) -> Tensor:
        x = functional.pad(x, self.padding, "reflect")
        return functional.conv2d(
            x, self.kernel.to(x.device), stride=self.stride, groups=self.in_channels
        )


def conv_padding(kernel_size: int, stride: int, dilation: int) -> int:
    return ((stride - 1) + dilation * (kernel_size - 1)) // 2



from pathlib import Path
import logging
import sys

from rich.logging import RichHandler
from torch import Tensor
from torch.nn import functional
from torch.utils.data import DataLoader
from torchinfo import summary
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms


sys.path.append("../src")


from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import MultilabelClassification

coco_labels = [
    "__background__",
    "person",
    "bicycle",
    "car",
    "motorbike",
    "aeroplane",
    "bus",
    "train",
    "truck",
    "boat",
    "trafficlight",
    "firehydrant",
    "streetsign",
    "stopsign",
    "parkingmeter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "hat",
    "backpack",
    "umbrella",
    "shoe",
    "eyeglasses",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sportsball",
    "kite",
    "baseballbat",
    "baseballglove",
    "skateboard",
    "surfboard",
    "tennisracket",
    "bottle",
    "plate",
    "wineglass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hotdog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "sofa",
    "pottedplant",
    "bed",
    "mirror",
    "diningtable",
    "window",
    "desk",
    "toilet",
    "door",
    "tvmonitor",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cellphone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "blender",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddybear",
    "hairdrier",
    "toothbrush",
    "hairbrush",
]

valid_coco_labels = coco_labels.copy()
for _ in [
    "__background__",
    "streetsign",
    "hat",
    "shoe",
    "eyeglasses",
    "plate",
    "mirror",
    "window",
    "desk",
    "door",
    "blender",
    "hairbrush",
]:
    valid_coco_labels.remove(_)


def collate_fn(batch):
    images = torch.stack([_[0] for _ in batch])
    labels = torch.stack(
        [
            functional.one_hot(
                torch.tensor(
                    [
                        valid_coco_labels.index(coco_labels[label])
                        for label in sample[1].get("labels", [])
                    ],
                    dtype=torch.int64,
                ).unique(),
                num_classes=len(valid_coco_labels),
            ).sum(dim=0)
            for sample in batch
        ]
    )
    return images, labels


class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int = 32) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data"

    def prepare_data(self) -> None:
        if not (self.data_dir / "coco2017").exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "awsaf49/coco-2017-dataset", path=self.data_dir, unzip=True, quiet=False
            )

    def setup(self, stage: str = "") -> None:
        coco_data_dir = self.data_dir / "coco2017"
        self.trainset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(coco_data_dir / "train2017"),
                str(coco_data_dir / "annotations" / "instances_train2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(800 - 1, max_size=800),
                        transforms.RandomCrop(800, pad_if_needed=True),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            ),
            target_keys=["labels"],
        )
        self.valset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(coco_data_dir / "val2017"),
                str(coco_data_dir / "annotations" / "instances_val2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(800 - 1, max_size=800),
                        transforms.RandomCrop(800, pad_if_needed=True),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            ),
            target_keys=["labels"],
        )

    def train_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="multilabel_classification"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_epochs=20,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=1,
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = MultilabelClassification(
            in_channels=backbone.out_channels, num_labels=len(valid_coco_labels)
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.SGD,
            optimizer_kwargs={"lr": 1e-2, "weight_decay": 1e-4, "momentum": 0.9},
            data_config={"categories": valid_coco_labels},
        )

    log.info(
        summary(
            model.model,
            input_data=model.model.backbone.dummy_input,
            row_settings=("var_names",),
            col_names=("output_size", "num_params", "mult_adds", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=16))



from typing import List

from torch import nn, Tensor

from sihl.layers.convblocks import ConvNormAct
from sihl.layers.scalers import BilinearScaler, AntialiasedDownscaler


class FPN(nn.Module):
    def __init__(
        self,
        in_channels: List[int],
        out_channels: int,
        bottom_level: int,
        top_level: int,
    ):
        super().__init__()
        assert 0 < bottom_level < top_level
        self.out_channels = (
            in_channels[:bottom_level]
            + [out_channels for _ in range(bottom_level, top_level + 1)]
            + in_channels[top_level:]
        )
        self.bottom_level, self.top_level = bottom_level, top_level
        self.lateral_convs = nn.ModuleList(
            ConvNormAct(in_c, out_channels, kernel_size=1)
            for in_c in in_channels[bottom_level : top_level + 1]
        )
        self.upscalers = nn.ModuleList(
            BilinearScaler(scale=2) for level in range(len(self.lateral_convs) - 1)
        )
        self.downscalers = nn.ModuleList(
            AntialiasedDownscaler(out_channels, out_channels)
            for _ in range(top_level + 1 - len(in_channels))
        )
        self.convs = nn.ModuleList(
            ConvNormAct(out_channels, out_channels)
            for _ in range(top_level - bottom_level + 1)
        )

    def forward(self, inputs: List[Tensor]) -> List[Tensor]:
        features = [
            lateral_conv(inputs[self.bottom_level + idx])
            for idx, lateral_conv in enumerate(self.lateral_convs)
        ]
        for idx, upscaler in enumerate(self.upscalers):
            features[-idx - 2] = upscaler(features[-idx - 1]) + features[-idx - 2]
        for downscaler in self.downscalers:
            features.append(downscaler(features[-1]))
        outputs = [conv(x) for conv, x in zip(self.convs, features)]
        return inputs[: self.bottom_level] + outputs + inputs[self.top_level + 1 :]



import os
import pytest
import tempfile

import torch

from sihl import TorchvisionBackbone, TORCHVISION_BACKBONE_NAMES


@pytest.mark.parametrize("name", TORCHVISION_BACKBONE_NAMES)
def test_torchvision_backbones(name: str) -> None:
    image = torch.rand((4, 3, 32, 32))
    with tempfile.TemporaryDirectory() as tmpdirname:
        os.environ["TORCH_HOME"] = tmpdirname
        model = TorchvisionBackbone(name)
    assert len(model.out_channels) == 6
    y = model(image)
    assert [_.shape[0] for _ in y] == [4 for _ in y]
    assert [_.shape[1] for _ in y] == model.out_channels
    assert [_.shape[2] for _ in y] == [32 // 2**idx for idx in range(len(y))]
    assert [_.shape[3] for _ in y] == [32 // 2**idx for idx in range(len(y))]


def test_grayscale_input() -> None:
    image = torch.rand((4, 1, 32, 32))
    model = TorchvisionBackbone("resnet18", input_channels=1)
    assert model.out_channels[0] == 1
    y = model(image[:, 0:1, :, :])
    assert y[0].shape[1] == 1


def test_multispectral_input() -> None:
    image = torch.rand((4, 9, 32, 32))
    model = TorchvisionBackbone("resnet18", input_channels=9)
    assert model.out_channels[0] == 9
    y = model(image)
    assert y[0].shape[1] == 9


def test_freeze_all_layers() -> None:
    TorchvisionBackbone("resnet18", frozen_levels=-1)


def test_freeze_some_layers() -> None:
    TorchvisionBackbone("resnet18", frozen_levels=3)


def test_top_level() -> None:
    image = torch.rand((4, 3, 2**7, 2**7))
    model = TorchvisionBackbone("resnet18", top_level=7)
    assert len(model.out_channels) == 8
    y = model(image)
    assert len(y) == 8
    assert y[7].shape[2] == y[7].shape[3] == 1
    assert y[6].shape[2] == y[6].shape[3] == 2


def test_unrecognized_backbone() -> None:
    with pytest.raises(ValueError):
        TorchvisionBackbone("resnet17")



from typing import Tuple, List, Dict

from torch import Tensor, nn
from torch.nn.functional import binary_cross_entropy_with_logits
from torchmetrics import MeanMetric, Accuracy, Precision, Recall
import torch

from sihl.layers import SequentialConvBlocks


class MultilabelClassification(nn.Module):
    """Multilabel classification is the prediction of the subset of all labels relevant
    to an input image. This subset can be empty if no label is relevant. In the
    particular case that the set of all possible labels is a singleton, we can call this
    task "binary classification"."""

    def __init__(
        self,
        in_channels: List[int],
        num_labels: int,
        num_channels: int = 256,
        num_layers: int = 1,
        level: int = 5,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_classes (int): Number of possible labels.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 1.
            level (int, optional): Level of inputs this head is attached to. Defaults to 5.
        """
        assert num_labels > 0, num_labels
        assert len(in_channels) > level, (len(in_channels), level)
        assert num_channels > 0 and num_layers > 0, (num_channels, num_layers)
        super().__init__()

        self.num_labels = num_labels
        self.level = level
        self.convs = nn.Sequential(
            SequentialConvBlocks(in_channels[level], num_channels, num_layers),
            nn.Conv2d(num_channels, num_labels, kernel_size=1),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
        )

        self.output_shapes = {
            "scores": ("batch_size", num_labels),
            "labels": ("batch_size", num_labels),
        }

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor]:
        scores, labels = torch.sort(
            torch.sigmoid(self.convs(inputs[self.level])), descending=True
        )
        return scores, labels

    def training_step(
        self, inputs: List[Tensor], target: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        logits = self.convs(inputs[self.level])
        loss = binary_cross_entropy_with_logits(logits, target.to(logits))
        return loss, {}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.accuracy_computer = Accuracy("multilabel", num_labels=self.num_labels)
        self.precision_computer = Precision("multilabel", num_labels=self.num_labels)
        self.recall_computer = Recall("multilabel", num_labels=self.num_labels)

    def validation_step(
        self, inputs: List[Tensor], target: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        logits = self.convs(inputs[self.level])
        target = target.to(logits)
        loss = binary_cross_entropy_with_logits(logits, target)
        self.loss_computer.to(loss.device).update(loss)
        self.accuracy_computer.to(logits.device).update(logits, target)
        self.precision_computer.to(logits.device).update(logits, target)
        self.recall_computer.to(logits.device).update(logits, target)
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "accuracy": self.accuracy_computer.compute().item(),
            "precision": self.precision_computer.compute().item(),
            "recall": self.recall_computer.compute().item(),
        }



from pathlib import Path
from typing import Tuple
import json
import logging

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision import tv_tensors
from torchvision.io import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import KeypointDetection
from sihl.layers import FPN


### HACK: torchvision doesn't support augmenting keypoints yet
# https://pytorch.org/blog/extending-torchvisions-transforms-to-object-detection-segmentation-and-video-tasks/#development-milestones-and-future-work
# So, we'll convert each keypoint to a tiny square bbox
# we'll augment those then convert them back by taking their center point
def polygons_to_bboxes(polygons: Tensor, canvas_size: Tuple[int, int]):
    """polygons: (N, K, 2)"""
    SIZE = 3  # bounding box size in pixels
    box_centers = polygons.reshape((-1, 2))
    flat_boxes = torch.cat([box_centers, torch.full_like(box_centers, SIZE)], dim=1)
    return tv_tensors.BoundingBoxes(
        flat_boxes, format=tv_tensors.BoundingBoxFormat.CXCYWH, canvas_size=canvas_size
    )


def bboxes_to_polygons(bboxes: tv_tensors.BoundingBoxes, num_vertices: int) -> Tensor:
    widths, heights = bboxes[:, 2], bboxes[:, 3]
    inside = ~((widths == 0) | (heights == 0)).reshape((-1, num_vertices))
    return bboxes[:, :2].reshape((-1, num_vertices, 2)), inside


###


class CocoHumanPoseEstimationDataset(torch.utils.data.Dataset):
    KEYPOINT_LABELS = [
        "nose",
        "left_eye",
        "right_eye",
        "left_ear",
        "right_ear",
        "left_shoulder",
        "right_shoulder",
        "left_elbow",
        "right_elbow",
        "left_wrist",
        "right_wrist",
        "left_hip",
        "right_hip",
        "left_knee",
        "right_knee",
        "left_ankle",
        "right_ankle",
    ]
    SKELETON = [
        ("left_ankle", "left_knee"),
        ("left_knee", "left_hip"),
        ("right_ankle", "right_knee"),
        ("right_knee", "right_hip"),
        ("left_hip", "right_hip"),
        ("left_shoulder", "left_hip"),
        ("right_shoulder", "right_hip"),
        ("left_shoulder", "right_shoulder"),
        ("left_shoulder", "left_elbow"),
        ("right_shoulder", "right_elbow"),
        ("left_elbow", "left_wrist"),
        ("right_elbow", "right_wrist"),
        ("left_eye", "right_eye"),
        ("nose", "left_eye"),
        ("nose", "right_eye"),
        ("left_eye", "left_ear"),
        ("right_eye", "right_ear"),
        ("left_ear", "left_shoulder"),
        ("right_ear", "right_shoulder"),
    ]

    def __init__(self, data_dir: Path, train: bool = False) -> None:
        self.image_size = 800
        self.train = train
        self.data_dir = data_dir
        self.annots_by_image = {}
        split = "train" if train else "val"
        self.image_dir = self.data_dir / f"{split}2017"

        with open(
            self.data_dir
            / "annotations_trainval2017"
            / f"person_keypoints_{split}2017.json"
        ) as f:
            coco_data = json.load(f)

        image_by_id = {
            image_annot["id"]: image_annot["file_name"]
            for image_annot in coco_data["images"]
        }
        for annot in coco_data["annotations"]:
            image_name = image_by_id[annot["image_id"]]
            if not (self.image_dir / image_name).exists():
                continue
            image_path = str((self.image_dir / image_name).resolve())
            if image_path not in self.annots_by_image:
                self.annots_by_image[image_path] = []
            self.annots_by_image[image_path].append(annot["keypoints"])
        self.annots_by_image = list(self.annots_by_image.items())
        self.transform = transforms.Compose(
            [
                transforms.Resize(self.image_size - 1, max_size=self.image_size),
                transforms.RandomCrop(self.image_size, pad_if_needed=True),
                transforms.ToDtype(torch.float32, scale=True),
            ]
        )

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, annot = self.annots_by_image[idx]
        image = torchvision.io.read_image(image_path, mode=ImageReadMode.RGB)
        target = torch.tensor(annot).reshape((-1, len(self.KEYPOINT_LABELS), 3))
        keypoints = target[:, :, :2].to(torch.float32)
        presence = target[:, :, 2].to(torch.bool)
        keypoints = polygons_to_bboxes(keypoints, image.shape[1:])
        image, keypoints = self.transform(image, keypoints)
        keypoints, inside = bboxes_to_polygons(keypoints, len(self.KEYPOINT_LABELS))
        presence = presence & inside
        keypoints = keypoints * presence.unsqueeze(2)
        # remove instances that have no present keypoint
        visible_instances = presence.any(dim=1)
        presence = presence[visible_instances]
        keypoints = keypoints[visible_instances]
        return image, {"keypoints": keypoints, "presence": presence}

    def __len__(self) -> int:
        return len(self.annots_by_image)

    @staticmethod
    def collate_fn(batch):
        images = torch.stack([_[0] for _ in batch])
        keypoints = [_[1]["keypoints"] for _ in batch]
        presence = [_[1]["presence"] for _ in batch]
        return images, {"keypoints": keypoints, "presence": presence}


class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int = 32) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "coco_2017"
        self.data_dir.mkdir(exist_ok=True)

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "clkmuhammed/microsoft-coco-2017-common-objects-in-context",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = CocoHumanPoseEstimationDataset(self.data_dir, train=True)
        self.validset = CocoHumanPoseEstimationDataset(self.data_dir, train=False)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            drop_last=True,
            collate_fn=CocoHumanPoseEstimationDataset.collate_fn,
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.validset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            drop_last=True,
            collate_fn=CocoHumanPoseEstimationDataset.collate_fn,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs",
        name="keypoint_detection",
        # version="tmp",
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_steps=90_000,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=1,
        precision="16-mixed",
        val_check_interval=1000,
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True, frozen_levels=1)
        neck = FPN(backbone.out_channels, 256, bottom_level=3, top_level=7)
        head = KeypointDetection(
            in_channels=neck.out_channels,
            num_keypoints=len(CocoHumanPoseEstimationDataset.KEYPOINT_LABELS),
            soft_label_decay_steps=90_000,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=torch.optim.SGD,
            optimizer_kwargs={"lr": 1e-2, "weight_decay": 1e-4, "momentum": 0.9},
            scheduler=torch.optim.lr_scheduler.MultiStepLR,
            scheduler_kwargs={
                "milestones": [60_000, 80_000],
                "gamma": 0.1,
                "warmup_batches": 1000,
            },
            data_config={
                "keypoints": CocoHumanPoseEstimationDataset.KEYPOINT_LABELS,
                "links": CocoHumanPoseEstimationDataset.SKELETON,
            },
        )

    log.debug(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=5,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=16))



from typing import List

from matplotlib import pyplot as plt
from torchvision.transforms.v2 import Resize
import numpy as np
import torch

from sihl.heads import DepthEstimation

from .common import get_images, plot_to_numpy


@get_images.register(DepthEstimation)
def _(head, config, input, target, features) -> List[np.ndarray]:
    prediction = head(features)
    target = target["targets"]  # TODO: masks
    resize = Resize(target[0].shape[1:], antialias=False)
    images = (resize(input).permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch, image in enumerate(images):
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.axis("off")
        axes[0].title.set_text("Input")
        axes[0].imshow(image)
        axes[1].title.set_text("Target")
        target_depth_map = target[batch].to("cpu")
        alpha = np.ones_like(target_depth_map)
        if "ignore" in config:
            alpha[target_depth_map == config["ignore"]] = 0
        axes[1].imshow(
            target_depth_map,
            cmap="Greys",
            vmin=target_depth_map.min(),
            vmax=target_depth_map.max(),
            alpha=alpha,
        )
        axes[2].title.set_text("Prediction")
        axes[2].imshow(
            prediction[batch].to("cpu"),
            cmap="Greys",
            vmin=target_depth_map.min(),
            vmax=target_depth_map.max(),
            alpha=alpha,
        )
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from typing import List, Tuple, Dict

from einops import rearrange, reduce
from torch import nn, Tensor
from torch.nn import functional
from torchmetrics import MeanMetric
from torchmetrics.regression import MeanAbsoluteError, MeanSquaredError
import torch

from sihl.heads.semantic_segmentation import SemanticSegmentation
from sihl.layers import SequentialConvBlocks
from sihl.utils import interpolate, EPS


class DepthEstimation(SemanticSegmentation):
    """Depth estimation is pixelwise regression.

    Refs:
        1. [Adabins](https://arxiv.org/abs/2011.14141)
    """

    def __init__(
        self,
        in_channels: List[int],
        lower_bound: float,
        upper_bound: float,
        bottom_level: int = 3,
        top_level: int = 5,
        num_channels: int = 256,
        num_layers: int = 1,
        num_bins: int = 256,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            lower_bound (float): Lower bound of the interval of possible values.
            upper_bound (float): Upper bound of the interval of possible values.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            num_bins (int, optional): Number of dynamic-sized bins. Defaults to 256.
        """
        assert lower_bound < upper_bound
        assert len(in_channels) > top_level >= bottom_level > 0
        assert num_channels > 0 and num_layers > 0
        assert num_bins > 1
        super().__init__(
            in_channels=in_channels,
            num_classes=num_bins,
            num_channels=num_channels,
            bottom_level=bottom_level,
            top_level=top_level,
            num_layers=num_layers,
        )

        self.num_bins = num_bins
        self.lower_bound, self.upper_bound = lower_bound, upper_bound
        self.bin_head = nn.Sequential(
            SequentialConvBlocks(in_channels[top_level], num_channels, num_layers),
            nn.Conv2d(num_channels, num_bins, kernel_size=1),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
        )
        stride = 2**bottom_level
        self.output_shapes = {
            "depth_maps": ("batch_size", f"height/{stride}", f"width/{stride}")
        }

    def normalize(self, x: Tensor) -> Tensor:
        return (x - self.lower_bound) / (self.upper_bound - self.lower_bound)

    def denormalize(self, x: Tensor) -> Tensor:
        return x * (self.upper_bound - self.lower_bound) + self.lower_bound

    def get_bin_centers(self, inputs: List[Tensor]) -> Tensor:
        bin_widths = self.bin_head(inputs[self.top_level]).relu() + EPS
        bin_widths = bin_widths / reduce(bin_widths, "n c -> n 1", "sum")
        return bin_widths.cumsum(dim=1) - bin_widths / 2

    def get_depth_map(self, inputs: List[Tensor], bin_centers: Tensor) -> Tensor:
        weights = self.get_logits(inputs).relu() + EPS
        weights = weights / reduce(weights, "n c h w -> n 1 h w", "sum")
        bin_centers = rearrange(bin_centers, "n c -> n c 1 1")
        depth_map = reduce(bin_centers * weights, "n c h w -> n 1 h w", "sum")
        return depth_map.clamp(0, 1)

    def forward(self, inputs: List[Tensor]) -> Tensor:
        bin_centers = self.get_bin_centers(inputs)
        depth_map = self.denormalize(self.get_depth_map(inputs, bin_centers))
        return interpolate(depth_map, size=inputs[0].shape[2:]).squeeze(1)

    def training_step(
        self, inputs: List[Tensor], targets: Tensor, masks: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        batch_size, target_height, target_width = targets.shape
        masks = rearrange(masks, "b h w -> b 1 h w")
        targets = rearrange(targets, "b h w -> b 1 h w")
        targets = self.normalize(targets)

        bin_centers = self.get_bin_centers(inputs)
        depth_map = self.get_depth_map(inputs, bin_centers=bin_centers)
        pred_shape = depth_map.shape[2:]
        depth_map = interpolate(depth_map, size=targets.shape[2:])

        g = (depth_map[masks] + EPS).log() - (targets[masks] + EPS).log()
        pix_loss = torch.sqrt(g.var() + 0.15 * g.mean().pow(2)) * 10

        masks = functional.interpolate(
            masks.to(torch.uint8), size=pred_shape, mode="nearest-exact"
        ).to(torch.bool)
        targets = interpolate(targets, size=pred_shape)
        hist_losses = []  # bidirectional chamfer loss
        for batch_idx in range(batch_size):
            target_hist = rearrange(targets[batch_idx][masks[batch_idx]], "k -> k 1")
            pred_hist = rearrange(bin_centers[batch_idx], "l -> 1 l")
            dist = (pred_hist - target_hist).pow(2)
            forward_chamfer = reduce(dist, "k l -> k", "min").mean()
            backward_chamfer = reduce(dist, "k l -> l", "min").mean()
            hist_losses.append(forward_chamfer + backward_chamfer)
        hist_loss = torch.stack(hist_losses).mean()
        return pix_loss + hist_loss, {"pixel_loss": pix_loss, "hist_loss": hist_loss}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.mae_computer = MeanAbsoluteError()
        self.rmse_computer = MeanSquaredError(squared=False)

    def validation_step(
        self, inputs: List[Tensor], targets: Tensor, masks: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        loss, _ = self.training_step(inputs, targets, masks)
        device = loss.device
        self.loss_computer.to(device).update(loss)
        depth_map = self.forward(inputs)
        batch_size, pred_height, pred_width = depth_map.shape
        self.rmse_computer.to(device).update(depth_map[masks], targets[masks])
        self.mae_computer.to(device).update(depth_map[masks], targets[masks])
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "rmse": self.rmse_computer.compute().item(),
            "mae": self.mae_computer.compute().item(),
        }



from typing import List

from matplotlib import patches
from matplotlib import pyplot as plt
from torch.nn import functional
import numpy as np
import torch

from sihl.heads import QuadrilateralDetection

from .common import get_images, plot_to_numpy, COLORS


@get_images.register(QuadrilateralDetection)
def _(
    head: QuadrilateralDetection, config, input, target, features
) -> List[np.ndarray]:
    categories = config["categories"] if "categories" in config else None
    prediction = head(features)
    if prediction is not None:
        num_instances, scores, pred_labels, pred_boxes = prediction
        saliency = head.get_saliency(features)
        saliency = functional.interpolate(
            saliency.unsqueeze(1), size=input.shape[2:], mode="nearest-exact"
        )
        saliency = saliency.squeeze(1).to("cpu").numpy()
    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch_idx, image in enumerate(images):
        seen_categories = []
        fig, axes = plt.subplots(1, 3, figsize=(10, 5), dpi=100)
        for ax in axes:
            ax.set_xticks([])
            ax.set_yticks([])
        axes[0].title.set_text("Input")
        axes[0].axis("off")
        axes[0].imshow(image)

        def get_patch(label, quad):
            label = str(label) if categories is None else categories[label]
            if label not in seen_categories:
                seen_categories.append(label)
                legend = label
            else:
                legend = None
            return patches.Polygon(
                quad,  # (4, 2)
                linewidth=1,
                edgecolor=[
                    _ / 255 for _ in COLORS[seen_categories.index(label) % len(COLORS)]
                ],
                facecolor="none",
                label=legend,
            )

        axes[1].title.set_text("Target")
        axes[1].imshow(np.full_like(image, fill_value=255))
        if target is not None:
            for label, quad in zip(
                target["classes"][batch_idx], target["quads"][batch_idx]
            ):
                axes[1].add_patch(get_patch(label.to("cpu"), quad.to("cpu")))
        axes[2].title.set_text("Prediction")
        axes[2].imshow(saliency[batch_idx], vmin=0, vmax=1)
        # np.full_like(image, fill_value=255))
        if prediction is not None:
            n = num_instances[batch_idx]
            for label, box in zip(
                pred_labels[batch_idx, :n], pred_boxes[batch_idx, :n]
            ):
                axes[2].add_patch(get_patch(label.to("cpu"), box.to("cpu")))
        fig.legend(loc="lower center", frameon=False, ncol=min(7, len(seen_categories)))
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



# ruff: noqa: F401
from sihl.layers.attention import SpatialAttention, ChannelAttention, CBAM, CrossCBAM
from sihl.layers.bifpn import BiFPN, FastNormalizedFusion
from sihl.layers.convblocks import SeparableConv2d, ConvNormAct, SequentialConvBlocks
from sihl.layers.fpn import FPN
from sihl.layers.pooling import BlurPool2d
from sihl.layers.preprocessing import Normalize, PadToMultipleOf
from sihl.layers.scalers import (
    AntialiasedDownscaler,
    BilinearAdditiveUpscaler,
    BilinearScaler,
    SimpleDownscaler,
    SimpleUpscaler,
    StridedDownscaler,
)



from functools import singledispatch
from io import BytesIO
from typing import List

from matplotlib.pyplot import Figure
import numpy as np

COLORS = (
    (0, 0, 0),
    (230, 25, 75),
    (60, 180, 75),
    (255, 225, 25),
    (0, 130, 200),
    (245, 130, 48),
    (145, 30, 180),
    (70, 240, 240),
    (240, 50, 230),
    (210, 245, 60),
    (250, 190, 212),
    (0, 128, 128),
    (220, 190, 255),
    (170, 110, 40),
    (255, 250, 200),
    (128, 0, 0),
    (170, 255, 195),
    (128, 128, 0),
    (255, 215, 180),
    (0, 0, 128),
    (128, 128, 128),
    (255, 255, 255),
)


def plot_to_numpy(fig: Figure) -> np.ndarray:
    io_buf = BytesIO()
    fig.savefig(io_buf, format="raw", dpi=100)
    io_buf.seek(0)
    img_arr = np.reshape(
        np.frombuffer(io_buf.getvalue(), dtype=np.uint8),
        newshape=(int(fig.bbox.bounds[3]), int(fig.bbox.bounds[2]), -1),
    )[:, :, :3]
    io_buf.close()
    return np.moveaxis(img_arr, 2, 0)  # HWC -> CHW


@singledispatch
def get_images(head, config, input, target, features) -> List[np.ndarray]:
    raise NotImplementedError(f'got unknown type "{type(head)}"')



# ruff: noqa: F401, F811
from torch import Tensor
import lightning.pytorch as pl
import pytest
import torch

from sihl import SihlLightningModule, SihlModel, TorchvisionBackbone, TimmBackbone
from sihl.heads import MulticlassClassification

from heads.test_multiclass_classification import model, targets


@pytest.fixture()
def torchvision_module(model: MulticlassClassification) -> SihlLightningModule:
    return SihlLightningModule(
        SihlModel(backbone=TorchvisionBackbone("resnet18"), neck=None, heads=[model])
    )


@pytest.fixture()
def timm_module(model: MulticlassClassification) -> SihlLightningModule:
    return SihlLightningModule(
        SihlModel(backbone=TimmBackbone("resnet18"), neck=None, heads=[model])
    )


@pytest.mark.parametrize("module", (torchvision_module, timm_module))
def test_forward_pass(module: SihlLightningModule, request) -> None:
    module = request.getfixturevalue(module.__name__)
    module(module.model.backbone.dummy_input)


@pytest.mark.parametrize("module", (torchvision_module, timm_module))
def test_training_step(module: SihlLightningModule, targets: Tensor, request) -> None:
    module = request.getfixturevalue(module.__name__)
    module.trainer = pl.Trainer()
    loss = module.training_step((torch.rand((4, 3, 32, 32)), [targets]), 0)
    assert loss.item()


@pytest.mark.parametrize("module", (torchvision_module, timm_module))
def test_validation_step(module: SihlLightningModule, targets: Tensor, request) -> None:
    module = request.getfixturevalue(module.__name__)
    module.trainer = pl.Trainer()
    module.on_validation_start()
    loss = module.validation_step((torch.rand((4, 3, 32, 32)), [targets]), 0)
    assert loss.item()
    module.on_validation_epoch_end()


@pytest.mark.parametrize("module", (torchvision_module, timm_module))
def test_configure_optimizers(module: SihlLightningModule, request) -> None:
    module = request.getfixturevalue(module.__name__)
    module.configure_optimizers()



from typing import List

from matplotlib import patches
from matplotlib import pyplot as plt
from torchvision.transforms.v2 import Resize
import numpy as np
import torch

from sihl.heads import KeypointDetection

from .common import get_images, plot_to_numpy, COLORS


def get_rectangle_patch(box, color):
    xy, width, height = (box[0], box[1]), box[2] - box[0], box[3] - box[1]
    return patches.Rectangle(
        xy, width, height, linewidth=1, edgecolor=color, facecolor="none"
    )


@get_images.register(KeypointDetection)
def _(head, config, input, target, features) -> List[np.ndarray]:
    keypoint_labels = config["keypoints"] if "keypoints" in config else []
    keypoint_links = config["links"] if "links" in config else []

    num_instances, scores, _, logits = head.forward_logits(features)
    batch_size, _, height, width = logits.shape
    heatmaps = logits.reshape(
        (batch_size, head.max_instances, head.num_kpts, height, width)
    )
    heatmaps = heatmaps.flatten(3, 4).softmax(dim=3).reshape(heatmaps.shape)
    keypoint_scores = heatmaps.flatten(3, 4).max(dim=3)[0]
    pred_keypoints = head.heatmap_to_keypoints(features, heatmaps)
    scores, reg_features = head.get_features(features)
    scores = scores.max(dim=2)[0]

    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch_idx, image in enumerate(images):
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.set_xticks([])
            ax.set_yticks([])
        axes[0].title.set_text("Input")
        axes[0].axis("off")
        axes[0].imshow(image)

        axes[1].title.set_text("Target")
        axes[1].imshow(image, alpha=0.2)
        if target is not None:
            for instance_idx in range(target["presence"][batch_idx].shape[0]):
                presence = target["presence"][batch_idx][instance_idx].to("cpu")
                keypoints = target["keypoints"][batch_idx][instance_idx].to("cpu")
                if keypoints.numel() > 0:
                    color = np.array(COLORS[instance_idx % len(COLORS)]) / 255
                    idxs = [idx for idx, pres in enumerate(presence) if pres is True]
                    xs = [keypoints[idx][0] + 0.5 for idx in idxs]
                    ys = [keypoints[idx][1] + 0.5 for idx in idxs]
                    axes[1].scatter(xs, ys, marker="o", s=3, color=color)
                    for link in keypoint_links:
                        try:
                            i = keypoint_labels.index(link[0])
                            j = keypoint_labels.index(link[1])
                        except ValueError:
                            continue
                        if presence[i] and presence[j]:
                            xs = [keypoints[i][0], keypoints[j][0]]
                            ys = [keypoints[i][1], keypoints[j][1]]
                            axes[1].plot(xs, ys, lw=2, c=color)

        axes[2].title.set_text("Prediction")
        total_hm = heatmaps[batch_idx, : target["presence"][batch_idx].shape[0]]
        total_hm = total_hm.sum(dim=(0, 1)).unsqueeze(0)
        resize = Resize(image.shape[:2], interpolation=0)
        axes[2].imshow(resize(total_hm).squeeze(0).to("cpu"))
        if pred_keypoints is not None:
            for instance_idx in range(target["presence"][batch_idx].shape[0]):
                presence = keypoint_scores[batch_idx, instance_idx].to("cpu") > 0.1
                keypoints = pred_keypoints[batch_idx, instance_idx].to("cpu")
                if keypoints.numel() > 0:
                    idxs = [idx for idx, pres in enumerate(presence) if pres is True]
                    xs = [keypoints[idx][0] + 0.5 for idx in idxs]
                    ys = [keypoints[idx][1] + 0.5 for idx in idxs]
                    axes[2].scatter(xs, ys, marker="o", s=3, color=color)
                    color = np.array(COLORS[instance_idx % len(COLORS)]) / 255
                    for link in keypoint_links:
                        try:
                            i = keypoint_labels.index(link[0])
                            j = keypoint_labels.index(link[1])
                        except ValueError:
                            continue
                        if presence[i] and presence[j]:
                            xs = [keypoints[i][0], keypoints[j][0]]
                            ys = [keypoints[i][1], keypoints[j][1]]
                            axes[1].plot(xs, ys, lw=2, c=color)

        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from typing import List

from einops import rearrange
from matplotlib import pyplot as plt
from torch import Tensor
import numpy as np
import torch

from sihl.heads import AnomalyDetection
from sihl.utils import interpolate

from .common import get_images, plot_to_numpy


def pca_reduce(x: Tensor, k: int) -> Tensor:
    N, C, H, W = x.shape
    x = x.view(N, C, -1)
    x_centered = x - x.mean(dim=(0, 2), keepdim=True)
    U, S, V = torch.svd(x_centered.transpose(1, 2))
    return (U[:, :, :k] @ torch.diag_embed(S[:, :k])).transpose(1, 2).view(N, k, H, W)


def normalized(x: Tensor) -> Tensor:
    return (x - x.min()) / (x.max() - x.min())


def to_rgb(x):
    return normalized(pca_reduce(x.unsqueeze(0), 3)[0].permute(1, 2, 0))


@get_images.register(AnomalyDetection)
def _(head, config, input, target, features) -> List[np.ndarray]:
    prediction = head(features)
    image = features[0]
    teacher_out = ((features[head.level] - head.features_mean) / head.feature_std).to(
        "cpu"
    )
    student_out = head.student(image)
    st_teacher = student_out[:, : head.out_channels].to("cpu")
    st_autoencoder = student_out[:, head.out_channels :].to("cpu")

    encoded = head.autoencoder_encoder(image)
    old_size = encoded.shape[2:]
    encoded = head.autoencoder_resize(encoded)
    h, w = encoded.shape[2:]
    encoded = rearrange(encoded, "b c h w -> b (c h w)")
    encoded = rearrange(
        head.autoencoder_bottleneck(encoded), "b (c h w) -> b c h w", h=h, w=w
    )
    ae_output = head.autoencoder_decoder(interpolate(encoded, size=old_size)).to("cpu")

    distance_st, distance_ae, distance_stae = head.compute_distances(features)
    local_anomaly = distance_st.mean(dim=1)
    global_anomaly = distance_stae.mean(dim=1)
    local_anomaly = (
        head.local_thresh
        * (local_anomaly - head.q_st_start)
        / (head.q_st_end - head.q_st_start)
    ).to("cpu")
    global_anomaly = (
        head.global_thresh
        * (global_anomaly - head.q_ae_start)
        / (head.q_ae_end - head.q_ae_start)
    ).to("cpu")

    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []

    for batch, image in enumerate(images):
        fig, axes = plt.subplots(nrows=3, ncols=3, figsize=(10, 10), dpi=100)
        for ax in axes:
            for subax in ax:
                subax.axis("off")
        axes[0][0].title.set_text("Input")
        axes[0][0].imshow(image)
        axes[0][1].title.set_text("Target")
        if target is not None:
            axes[0][1].imshow(target[batch].to("cpu"), cmap="gray")
        axes[0][2].title.set_text("Prediction")
        if prediction is not None:
            axes[0][2].imshow(prediction[batch].to("cpu"), cmap="gray", vmin=0, vmax=1)

        axes[1][0].title.set_text("autoencoder")
        axes[1][0].imshow(to_rgb(ae_output[batch]))
        axes[1][1].title.set_text("student (autoencoder)")
        axes[1][1].imshow(to_rgb(st_autoencoder[batch]))
        axes[1][2].title.set_text("Global anomaly")
        axes[1][2].imshow(global_anomaly[batch], cmap="seismic", vmin=-1, vmax=1)

        axes[2][0].title.set_text("teacher")
        axes[2][0].imshow(to_rgb(teacher_out[batch]))
        axes[2][1].imshow(to_rgb(st_teacher[batch]))
        axes[2][1].title.set_text("student (teacher)")
        axes[2][2].title.set_text("Local anomaly")
        axes[2][2].imshow(local_anomaly[batch], cmap="seismic", vmin=-1, vmax=1)

        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from typing import Any, Tuple, List, Dict

from einops import rearrange, repeat, reduce
from torch import nn, Tensor
from torchmetrics import MeanMetric
from torchmetrics.detection.mean_ap import MeanAveragePrecision
from torchvision.ops import box_iou, complete_box_iou_loss
import torch

from sihl.layers import SequentialConvBlocks, ConvNormAct
from sihl.utils import coordinate_grid, f_score, EPS, focal_loss, interpolate


class ObjectDetection(nn.Module):
    """Object detection is the prediction of the set of "objects" (pairs of axis-aligned
    rectangular bounding boxes and the corresponding category) in the input image.

    Refs:
        1. [One-to-Few Label Assignment for End-to-End Dense Detection](https://arxiv.org/abs/2303.11567)
        2. [TOOD: Task-aligned One-stage Object Detection](https://arxiv.org/abs/2108.07755)
    """

    def __init__(
        self,
        in_channels: List[int],
        num_classes: int,
        bottom_level: int = 3,
        top_level: int = 7,
        num_channels: int = 256,
        num_layers: int = 4,
        max_instances: int = 100,
        t_min: float = 0.2,
        t_max: float = 0.6,
        topk: int = 7,
        soft_label_decay_steps: int = 1,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_classes (int): Number of possible object categories.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            max_instances (int, optional): Maximum number of instances to predict in a sample. Defaults to 100.
            t_min (float, optional): Lower bound of O2F parameter t. Defaults to 0.2.
            t_max (float, optional): Upper bound of O2F parameter t. Defaults to 0.6.
            topk (int, optional): How many anchors to match to each ground truth object when copmuting the loss. Defaults to 7.
            soft_label_decay_steps (int, optional): How many training steps to perform before the one-to-few matching becomes one-to-one. Defaults to 1.
        """
        assert num_classes > 0, num_classes
        assert len(in_channels) > top_level, (len(in_channels), top_level)
        assert 0 < bottom_level <= top_level, (bottom_level, top_level)
        assert num_channels > 4 and num_layers > 0, (num_channels, num_layers)
        assert max_instances > 0, max_instances
        assert 0 <= t_min <= t_max <= 1, (t_min, t_max)
        assert topk > 0 and soft_label_decay_steps > 0, (topk, soft_label_decay_steps)
        super().__init__()

        self.in_channels = in_channels
        self.num_classes = num_classes
        self.bottom_level = bottom_level
        self.top_level = top_level
        self.levels = range(bottom_level, top_level + 1)
        self.num_channels = num_channels
        self.num_layers = num_layers
        self.max_instances = max_instances
        self.t_min, self.t_max, self.topk = t_min, t_max, topk
        self.soft_label_decay_steps, self.current_step = soft_label_decay_steps, 0
        self.register_buffer("threshold", torch.tensor(0.5))  # adjusted in validation
        self.box_iou = box_iou  # enables overriding these in children classes
        self.box_loss = complete_box_iou_loss

        # input channels need to be matched if they aren't already
        matched_in_channels = in_channels[bottom_level]
        self.lateral_convs = None
        if not all(in_channels[lvl] == matched_in_channels for lvl in self.levels):
            matched_in_channels = num_channels
            self.lateral_convs = nn.ModuleList(
                [ConvNormAct(in_channels[lvl], num_channels, 1) for lvl in self.levels]
            )

        self.regression_stem = SequentialConvBlocks(
            matched_in_channels, num_channels, num_layers=num_layers, norm="group"
        )
        self.box_head = ConvNormAct(num_channels, 4, 1, norm=None, act="softplus")
        self.class_head = nn.Sequential(
            SequentialConvBlocks(
                matched_in_channels, num_channels, num_layers=num_layers, norm="group"
            ),
            ConvNormAct(num_channels, num_classes, 1, norm=None, act="sigmoid"),
        )
        self.centerness_head = nn.Sequential(
            SequentialConvBlocks(
                num_channels, num_channels // 4, num_layers=num_layers, norm="group"
            ),
            ConvNormAct(num_channels // 4, 1, kernel_size=1, norm=None, act="sigmoid"),
        )
        self.output_shapes = {
            "num_instances": ("batch_size",),
            "scores": ("batch_size", max_instances),
            "classes": ("batch_size", max_instances),
            "boxes": ("batch_size", max_instances, 4),
        }

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        scores, reg_features = self.get_features(inputs)
        scores, pred_classes = scores.max(dim=2)
        pred_boxes = self.get_boxes(reg_features)
        batches = repeat(torch.arange(scores.shape[0]), f"b -> b {self.max_instances}")
        instance_idxs = scores.topk(self.max_instances).indices
        scores = scores[batches, instance_idxs]
        pred_classes = pred_classes[batches, instance_idxs]
        pred_boxes = pred_boxes[batches, instance_idxs]
        num_instances = reduce(scores >= self.threshold, "b i -> b", "sum")
        return num_instances, scores, pred_classes, pred_boxes

    def get_features(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor]:
        torch._assert(len(inputs) > self.top_level, "too few input levels")
        xs = [inputs[level] for level in self.levels]
        if self.lateral_convs is not None:
            xs = [lat(x) for lat, x in zip(self.lateral_convs, xs)]
        reg_features = [self.regression_stem(x) for x in xs]
        centers = [self.centerness_head(reg_feat) for reg_feat in reg_features]
        scores = [self.class_head(x) for x in xs]
        scores = [
            rearrange(s * c, "b c h w -> b (h w) c") for s, c in zip(scores, centers)
        ]
        scores = torch.cat(scores, dim=1)
        return scores, reg_features

    def get_saliency(self, inputs: List[Tensor]) -> Tensor:
        torch._assert(len(inputs) > self.top_level, "too few input levels")
        xs = [inputs[level] for level in self.levels]
        if self.lateral_convs is not None:
            xs = [lat(x) for lat, x in zip(self.lateral_convs, xs)]
        reg_features = [self.regression_stem(x) for x in xs]
        centers = [self.centerness_head(x) for x in reg_features]
        scores = [self.class_head(x) for x in xs]
        size = xs[0].shape[2:]
        scores = [
            interpolate(reduce(s * c, "b c h w -> b 1 h w", "max"), size=size)
            for s, c in zip(scores, centers)
        ]
        return reduce(torch.cat(scores, dim=1), "b c h w -> b h w", "mean")

    def get_boxes(self, features: List[Tensor]) -> Tensor:
        boxes = []
        for level, x in zip(self.levels, features):
            stride, (h, w) = 2**level, x.shape[2:]
            biases = coordinate_grid(h, w, h * stride, w * stride).to(x)
            biases = repeat(biases, "c h w -> (2 c) h w", c=2)
            biases = rearrange(biases, "c h w -> 1 (h w) c", c=4)
            level_boxes = rearrange(self.box_head(x), "b c h w -> b (h w) c", c=4)
            scale = stride * torch.tensor([[[-1, -1, 1, 1]]]).to(x)
            boxes.append(biases + level_boxes * scale)
        return torch.cat(boxes, dim=1)

    def inside_box(self, points: Tensor, box: Tensor) -> Tensor:
        mask = (points[:, 0] > box[0]) & (points[:, 0] < box[2])
        return mask & (points[:, 1] > box[1]) & (points[:, 1] < box[3])

    @torch.no_grad()
    def get_targets(
        self,
        anchors: Tensor,
        scores: Tensor,
        pred_boxes: Tensor,
        gt_classes: List[Tensor],
        gt_boxes: List[Tensor],
        is_validating: bool = False,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        progress = min(max(0, self.current_step / self.soft_label_decay_steps), 1)
        ambiguous_samples_factor = (self.t_min - self.t_max) * progress + self.t_max
        if not is_validating:
            self.current_step += 1
        (batch_size, num_instances, _), device = scores.shape, scores.device
        class_target = torch.zeros_like(scores)
        target_shape = (batch_size, num_instances)
        assignment = torch.zeros(target_shape, device=device, dtype=torch.long)
        box_target = torch.zeros(target_shape + pred_boxes.shape[2:], device=device)
        for batch_idx in range(batch_size):
            for gt_idx in range(gt_boxes[batch_idx].shape[0]):
                gt_box = gt_boxes[batch_idx][gt_idx]
                gt_class = gt_classes[batch_idx][gt_idx]
                # only consider predictions whose anchor lie inside the gt box
                mask = self.inside_box(anchors, gt_box)
                mask = mask.nonzero(as_tuple=True)[0].to(device)
                if mask.numel() == 0:  # can happen if the gt box is too small
                    continue
                candidate_ious = self.box_iou(
                    gt_box.unsqueeze(0), pred_boxes[batch_idx, mask]
                ).squeeze(0)
                candidate_scores = scores[batch_idx, mask, gt_class]
                with torch.autocast(device_type="cuda", enabled=False):
                    candidate_scores = candidate_scores.to(torch.float32)
                    quality = candidate_scores.pow(0.2) * candidate_ious.pow(0.8)
                quality = quality.nan_to_num(0).to(scores)  # where are the nans from?
                topk_values, topk_idxs = quality.topk(k=min(quality.numel(), self.topk))
                if len(topk_values) > 1 and topk_values[1] > EPS:
                    topk_values = topk_values.to(scores) / topk_values[1]
                    topk_values = topk_values * ambiguous_samples_factor
                topk_values[0] = 1.0  # "certain" anchor is at index 0 (it's sorted)
                topk_idxs = mask[topk_idxs]
                for pos_idx, pos_value in zip(topk_idxs, topk_values):
                    if pos_value < max(EPS, class_target[batch_idx, pos_idx].max()):
                        continue
                    class_target[batch_idx, pos_idx, gt_class] = pos_value
                    assignment[batch_idx, pos_idx] = gt_idx
                    box_target[batch_idx, pos_idx] = gt_box
        return class_target, box_target, assignment

    def get_anchors(self, inputs: List[Tensor], normalized: bool = False) -> Tensor:
        anchors_by_level = []
        for level in self.levels:
            height, width = inputs[level].shape[2:]
            y_max = 1.0 if normalized else height * 2**level
            x_max = 1.0 if normalized else width * 2**level
            coords = coordinate_grid(height, width, y_max, x_max).to(inputs[level])
            anchors_by_level.append(rearrange(coords, "c h w -> (h w) c", c=2))
        return torch.cat(anchors_by_level)

    def get_losses(
        self,
        scores: Tensor,
        pred_boxes: Tensor,
        class_target: Tensor,
        box_target: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        pos_mask = reduce(class_target, "b i s -> b i", "max") > EPS
        if not pos_mask.any():
            class_loss = torch.zeros(1, device=scores.device, requires_grad=True)
            box_loss = torch.zeros(1, device=scores.device, requires_grad=True)
            return class_loss, box_loss
        with torch.autocast(device_type="cuda", enabled=False):
            scores = scores.to(torch.float32).nan_to_num(nan=0.0)
            class_target = class_target.to(torch.float32)
            class_loss = focal_loss(scores, class_target).sum() / class_target.sum()
            box_target = box_target.to(torch.float32)
            box_loss = self.box_loss(
                pred_boxes[pos_mask], box_target[pos_mask], reduction="mean"
            )
        return class_loss, box_loss

    def training_step(
        self,
        inputs: List[Tensor],
        classes: List[Tensor],
        boxes: List[Tensor],
        is_validating: bool = False,
    ) -> Tuple[Tensor, Dict[str, Any]]:
        scores, reg_features = self.get_features(inputs)
        pred_boxes = self.get_boxes(reg_features)
        class_target, box_target, assignment = self.get_targets(
            self.get_anchors(inputs), scores, pred_boxes, classes, boxes, is_validating
        )
        class_loss, box_loss = self.get_losses(
            scores, pred_boxes, class_target, box_target
        )
        return class_loss + box_loss, {"class_loss": class_loss, "box_loss": box_loss}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.map_computer = MeanAveragePrecision(
            extended_summary=True, backend="faster_coco_eval"
        )

    def validation_step(
        self, inputs: List[Tensor], classes: List[Tensor], boxes: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        num_instances, scores, pred_classes, pred_boxes = self.forward(inputs)
        self.map_computer.to(pred_boxes.device).update(
            [
                {"scores": s, "labels": c, "boxes": b}
                for s, c, b in zip(scores, pred_classes, pred_boxes)
            ],
            [{"labels": c, "boxes": b} for c, b in zip(classes, boxes)],
        )
        loss, metrics = self.training_step(inputs, classes, boxes, is_validating=True)
        self.loss_computer.to(loss.device).update(loss)
        return loss, metrics

    def on_validation_end(self) -> Dict[str, Any]:
        metrics = self.map_computer.compute()
        precision = metrics["precision"][0, :, :, 0, 2].mean(dim=1)
        recall = torch.as_tensor(self.map_computer.rec_thresholds)
        scores = metrics["scores"][0, :, :, 0, 2].mean(dim=1)
        f1 = f_score(1.0)(precision, recall)
        best_idx = int(f1.argmax().item())
        self.threshold = scores[best_idx]
        metrics["threshold"] = self.threshold
        metrics["precision"] = precision[best_idx]
        metrics["recall"] = recall[best_idx]
        metrics["f1"] = f1[best_idx]
        metrics["f0.5"] = f_score(0.5)(precision[best_idx], recall[best_idx])
        metrics["f2"] = f_score(2.0)(precision[best_idx], recall[best_idx])
        metrics["loss"] = self.loss_computer.compute()
        del metrics["map_per_class"], metrics["mar_100_per_class"]
        del metrics["scores"], metrics["classes"], metrics["ious"]
        return {k: v.to(self.map_computer.device) for k, v in metrics.items()}



from typing import List

from matplotlib import patches
from matplotlib import pyplot as plt
import numpy as np
import torch

from sihl.heads import ObjectDetection

from .common import get_images, plot_to_numpy, COLORS


@get_images.register(ObjectDetection)
def _(head, config, input, target, features) -> List[np.ndarray]:
    categories = config["categories"] if "categories" in config else None
    prediction = head(features)
    if prediction is not None:
        pred_labels, pred_boxes = prediction[2], prediction[3]
    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch, image in enumerate(images):
        seen_categories = []
        fig, axes = plt.subplots(1, 3, figsize=(10, 5), dpi=100)
        for ax in axes:
            ax.set_xticks([])
            ax.set_yticks([])
        axes[0].title.set_text("Input")
        axes[0].axis("off")
        axes[0].imshow(image)

        def get_patch(label, box):
            label = str(label) if categories is None else categories[label]
            if label not in seen_categories:
                seen_categories.append(label)
                legend = label
            else:
                legend = None
            return patches.Rectangle(
                (box[0], box[1]),
                box[2] - box[0],
                box[3] - box[1],
                linewidth=1,
                edgecolor=[
                    _ / 255 for _ in COLORS[seen_categories.index(label) % len(COLORS)]
                ],
                facecolor="none",
                label=legend,
            )

        axes[1].title.set_text("Target")
        axes[1].imshow(np.full_like(image, fill_value=255))
        if target is not None:
            for label, box in zip(target["classes"][batch], target["boxes"][batch]):
                axes[1].add_patch(get_patch(label.to("cpu"), box.to("cpu")))
        axes[2].title.set_text("Prediction")
        axes[2].imshow(np.full_like(image, fill_value=255))
        if prediction is not None:
            n = prediction[0][batch]
            for label, box in zip(pred_labels[batch][:n], pred_boxes[batch][:n]):
                axes[2].add_patch(get_patch(label.to("cpu"), box.to("cpu")))
        fig.legend(loc="lower center", frameon=False, ncol=min(7, len(seen_categories)))
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from typing import List

from matplotlib import pyplot as plt
from torch.nn import functional
import numpy as np
import torch

from sihl.heads import MetricLearning

from .common import get_images, plot_to_numpy


@get_images.register(MetricLearning)
def _(head, config, input, target, features) -> List[np.ndarray]:
    embeddings = head(features)
    similarities = functional.linear(embeddings, head.index_embeddings)
    top5_values, top5_labels = torch.topk(similarities, k=6)
    top5_labels = head.index_ids[top5_labels][:, 1:6].cpu().numpy()  # argmax?
    top5_scores = top5_values[:, 1:6].cpu().numpy()
    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch, image in enumerate(images):
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.axis("off")
        axes[0].title.set_text("Input")
        axes[0].imshow(image)

        def set_text(ax, text):
            ax.text(
                0.5,
                0.5,
                text,
                transform=ax.transAxes,
                ha="center",
                va="center",
                fontsize="large",
            )

        axes[1].title.set_text("Target")
        axes[1].imshow(np.full_like(image, fill_value=255))
        if target is not None:
            set_text(axes[1], "id: " + str(target[batch].item()))
        axes[2].title.set_text("Prediction")
        axes[2].imshow(np.full_like(image, fill_value=255))
        if embeddings is not None:
            patch_text = ""
            for rank, (score, label) in enumerate(
                zip(top5_scores[batch], top5_labels[batch])
            ):
                patch_text += (
                    f"{rank+1}. id: {str(label)} ({int(score*100)} % similar)\n"
                )
            set_text(axes[2], patch_text.strip())
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from typing import Tuple, List, Union, Dict

from torch import nn, Tensor
from torch.nn import functional
from torchmetrics import JaccardIndex, MeanMetric, Accuracy
import torch

from sihl.layers import (
    ConvNormAct,
    SequentialConvBlocks,
    SimpleUpscaler,
    BilinearScaler,
)
from sihl.utils import interpolate


class SemanticSegmentation(nn.Module):
    """Semantic segmentation is pixelwise multiclass classification.

    Refs:
        1. [PP-LiteSeg](https://arxiv.org/abs/2204.02681)
    """

    def __init__(
        self,
        in_channels: List[int],
        num_classes: int,
        bottom_level: int = 3,
        top_level: int = 5,
        num_channels: int = 256,
        num_layers: int = 3,
        pool_sizes: List[int] = [1, 2, 4],
        ignore_index: Union[int, None] = None,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_classes (int): Number of possible pixel categories.
            bottom_level (int, optional): Bottom level of inputs this head is attached to. Defaults to 3.
            top_level (int, optional): Top level of inputs this head is attached to. Defaults to 7.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
            ignore_index (Union[int, None], optional): Ignored category index. Defaults to None.
        """
        assert num_classes > 0
        assert len(in_channels) > top_level >= bottom_level > 0
        assert num_channels > 0 and num_layers >= 0
        super().__init__()

        self.in_channels = in_channels
        self.num_classes = num_classes
        self.num_channels = num_channels
        self.num_layers = num_layers
        self.pool_sizes = tuple(pool_sizes)
        self.bottom_level = bottom_level
        self.top_level = top_level
        self.ignore_index = ignore_index or -100
        self.levels = list(range(bottom_level, top_level + 1))
        self.rev_levels = list(reversed(range(bottom_level, top_level)))
        self.context_aggregation = SPPM(
            in_channels[top_level], num_channels, self.pool_sizes
        )
        self.lateral_convs = nn.ModuleList(
            [ConvNormAct(in_channels[level], num_channels) for level in self.rev_levels]
        )
        self.upscalers = nn.ModuleList(
            [SimpleUpscaler(num_channels, num_channels) for level in self.rev_levels]
        )
        self.fusions = nn.ModuleList(
            [UAFM(num_channels, num_channels) for level in self.rev_levels]
        )
        self.out_conv = nn.Sequential(
            SequentialConvBlocks(num_channels, num_channels, num_layers),
            nn.Conv2d(num_channels, num_classes, kernel_size=1),
        )
        self.output_shapes = {
            "score_maps": ("batch_size", "height", "width"),
            "class_maps": ("batch_size", "height", "width"),
        }

    def get_logits(self, inputs: List[Tensor]) -> Tensor:
        x = self.context_aggregation(inputs[self.top_level])
        for level, lateral, upscale, fuse in zip(
            self.rev_levels, self.lateral_convs, self.upscalers, self.fusions
        ):
            x = fuse(lateral(inputs[level]), upscale(x))
        return self.out_conv(x)

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor]:
        x = interpolate(self.get_logits(inputs), size=inputs[0].shape[2:])
        return x.softmax(dim=1).max(dim=1)

    def training_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        logits = interpolate(self.get_logits(inputs), size=targets.shape[1:])
        loss = functional.cross_entropy(logits, targets, ignore_index=self.ignore_index)
        return loss, {}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        metric_kwargs = {
            "task": "multiclass",
            "num_classes": self.num_classes,
            "ignore_index": self.ignore_index,
        }
        self.pixel_accuracy = Accuracy(**metric_kwargs)
        self.mean_iou_computer = JaccardIndex(**metric_kwargs)

    def validation_step(
        self, inputs: List[Tensor], targets: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        logits = interpolate(self.get_logits(inputs), size=targets.shape[1:])
        loss = functional.cross_entropy(logits, targets, ignore_index=self.ignore_index)
        scores = logits.softmax(dim=1)
        self.mean_iou_computer.to(loss.device).update(scores, targets)
        self.pixel_accuracy.to(loss.device).update(scores, targets)
        self.loss_computer.to(loss.device).update(loss)
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "pixel_accuracy": self.pixel_accuracy.compute().item(),
            "mean_iou": self.mean_iou_computer.compute().item(),
        }


class SPPM(nn.Module):
    """https://arxiv.org/abs/2204.02681"""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        pool_sizes: Tuple[int] = (1, 2, 4),
        with_shortcut: bool = False,
    ) -> None:
        super().__init__()
        self.with_shortcut = with_shortcut
        self.pools = nn.ModuleList(
            [
                nn.Sequential(
                    # HACK: `AdaptiveAvgPool2d` fails ONNX export, so use interpolation
                    BilinearScaler(size=pool_size),
                    ConvNormAct(in_channels, out_channels, 1),
                )
                for pool_size in pool_sizes
            ]
            if len(pool_sizes) > 0
            else [nn.Identity()]
        )
        if with_shortcut:
            self.shortcut = ConvNormAct(in_channels, out_channels, 1)
        self.out_conv = ConvNormAct(out_channels, out_channels, 1)

    def forward(self, x: Tensor) -> Tensor:
        fused = torch.stack(
            [interpolate(pool(x), size=x.shape[2:]) for pool in self.pools]
        ).sum(0)
        if self.with_shortcut:
            fused = fused + self.shortcut(x)
        return self.out_conv(fused)


class UAFM(nn.Module):
    """https://arxiv.org/abs/2204.02681"""

    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.conv = ConvNormAct(4, 1, norm=None, act="sigmoid")

    def forward(self, x1: Tensor, x2: Tensor) -> Tensor:
        alpha = self.conv(
            torch.cat(
                [
                    torch.mean(x1, dim=1, keepdim=True),
                    torch.max(x1, dim=1, keepdim=True).values,
                    torch.mean(x2, dim=1, keepdim=True),
                    torch.max(x2, dim=1, keepdim=True).values,
                ],
                dim=1,
            )
        )
        return x1 * alpha + x2 * (1 - alpha)



from pathlib import Path
from typing import List

from torch import Tensor
import numpy as np
import onnx
import onnxruntime
import pytest
import torch

from sihl.heads import ViewInvarianceLearning


BATCH_SIZE, NUM_CHANNELS, HEIGHT, WIDTH = 4, 256, 128, 128
EMBEDDING_DIM = 512
NUM_CLASSES, LEVEL = 100, 5
ONNX_VERSION, ONNX_FILE_NAME = 18, "view_invariance_learning.onnx"


@pytest.fixture()
def model() -> ViewInvarianceLearning:
    return ViewInvarianceLearning(
        in_channels=[3] + [NUM_CHANNELS for _ in range(LEVEL)],
        embedding_dim=EMBEDDING_DIM,
    )


@pytest.fixture()
def backbone_output() -> List[Tensor]:
    return [torch.randn((BATCH_SIZE, 3, HEIGHT, WIDTH))] + [
        torch.randn((BATCH_SIZE, NUM_CHANNELS, HEIGHT // 2**_, WIDTH // 2**_))
        for _ in range(LEVEL)
    ]


def test_forward(model: ViewInvarianceLearning, backbone_output: List[Tensor]) -> None:
    embeddings = model(backbone_output)
    assert tuple(embeddings.shape) == (BATCH_SIZE, EMBEDDING_DIM)


def test_training_step(
    model: ViewInvarianceLearning, backbone_output: List[Tensor]
) -> None:
    loss, _ = model.training_step(backbone_output, backbone_output)
    assert loss.item()


def test_validation_step(
    model: ViewInvarianceLearning, backbone_output: List[Tensor]
) -> None:
    model.on_validation_start()
    loss, _ = model.validation_step(backbone_output, backbone_output)
    assert loss.item()
    metrics = model.on_validation_end()
    assert metrics


@pytest.fixture()
def onnx_model(model: ViewInvarianceLearning, backbone_output: List[Tensor]) -> None:
    torch.onnx.export(
        model.to(torch.float32),
        args=backbone_output,
        f=ONNX_FILE_NAME,
        opset_version=ONNX_VERSION,
        input_names=[f"input_level_{idx}" for idx in range(len(backbone_output))],
        output_names=[f"head0/{name}" for name, shape in model.output_shapes.items()],
        dynamic_axes=dict(
            {
                f"input_level_{lvl}": {
                    0: "batch_size",
                    2: f"height/{2**lvl}",
                    3: f"width/{2**lvl}",
                }
                for lvl in range(len(backbone_output))
            },
            **{
                f"head0/{name}": {
                    shape_idx: str(shape_value)
                    for shape_idx, shape_value in enumerate(shape)
                }
                for name, shape in model.output_shapes.items()
            },
        ),
        external_data=False,
        verify=True,
        # dynamo=True,
        # report=True,
    )
    onnx_model = onnx.load(ONNX_FILE_NAME)
    Path(ONNX_FILE_NAME).unlink()
    return onnx_model


def test_onnx_inference(
    onnx_model, model: ViewInvarianceLearning, backbone_output: List[Tensor]
) -> None:
    model.eval()
    onnx_session = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    onnx_input = {
        f"input_level_{idx}": _.numpy()
        for idx, _ in enumerate(backbone_output)
        if f"input_level_{idx}" in [node.name for node in onnx_model.graph.input]
    }
    # just check that 99% of values are equal.
    pytorch_output = model(backbone_output).detach().numpy()
    onnx_output = onnx_session.run(None, onnx_input)[0]
    assert np.sum(np.abs(pytorch_output - onnx_output) > 1) / onnx_output.size < 0.01



from typing import List, Tuple, Dict

from torch import Tensor, nn
from torch.nn.functional import cross_entropy
from torchmetrics import MeanMetric, Accuracy, Precision, Recall
import torch

from sihl.layers import SequentialConvBlocks


class MulticlassClassification(nn.Module):
    """Multiclass classification is the prediction of the most probable category (or
    "class") for an input image among a finite set of possible categories. One input
    is always matched to exactly one category.
    In the particular case where the set of possible categories is totally ordered, we
    can call this task "ordinal classification"."""

    def __init__(
        self,
        in_channels: List[int],
        num_classes: int,
        num_channels: int = 256,
        num_layers: int = 1,
        level: int = 5,
        label_smoothing: float = 0.0,
        is_ordinal: bool = False,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_classes (int): Number of possible categories.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 1.
            level (int, optional): Level of inputs this head is attached to. Defaults to 5.
            label_smoothing (float, optional): How much to smooth target probability distribution. Defaults to 0.0.
            is_ordinal (bool, optional): Whether the categories are ordered. Defaults to False.
        """
        assert num_classes > 0, num_classes
        assert len(in_channels) > level, (len(in_channels), level)
        assert num_channels > 0 and num_layers > 0, (num_channels, num_layers)
        super().__init__()

        self.num_classes = num_classes
        self.level = level
        self.label_smoothing = label_smoothing
        self.is_ordinal = is_ordinal
        self.convs = nn.Sequential(
            SequentialConvBlocks(in_channels[level], num_channels, num_layers),
            nn.Conv2d(num_channels, num_classes, kernel_size=1),
            nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
        )

        self.output_shapes = {
            "scores": ("batch_size", num_classes),
            "classes": ("batch_size",),
        }

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor]:
        scores, classes = self.convs(inputs[self.level]).softmax(dim=1).max(dim=1)
        return scores, classes

    def training_step(
        self, inputs: List[Tensor], target: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        logits = self.convs(inputs[self.level])
        target = target.to(logits.device)
        if self.is_ordinal:
            target = soft_ordinal_category(target, self.num_classes)
        loss = cross_entropy(logits, target, label_smoothing=self.label_smoothing)
        return loss, {}

    def on_validation_start(self) -> None:
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.accuracy_computer = Accuracy("multiclass", num_classes=self.num_classes)
        self.precision_computer = Precision("multiclass", num_classes=self.num_classes)
        self.recall_computer = Recall("multiclass", num_classes=self.num_classes)

    def validation_step(
        self, inputs: List[Tensor], target: Tensor
    ) -> Tuple[Tensor, Dict[str, float]]:
        logits = self.convs(inputs[self.level])
        target = target.to(logits.device)
        loss = cross_entropy(logits, target, label_smoothing=self.label_smoothing)
        self.loss_computer.to(loss.device).update(loss)
        self.accuracy_computer.to(logits.device).update(logits, target)
        self.precision_computer.to(logits.device).update(logits, target)
        self.recall_computer.to(logits.device).update(logits, target)
        return loss, {}

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "accuracy": self.accuracy_computer.compute().item(),
            "precision": self.precision_computer.compute().item(),
            "recall": self.recall_computer.compute().item(),
        }


def soft_ordinal_category(
    labels: Tensor, num_labels: int, peakiness: float = 1.0
) -> Tensor:
    """https://openaccess.thecvf.com/content_CVPR_2019/papers/Diaz_Soft_Labels_for_Ordinal_Regression_CVPR_2019_paper.pdf"""
    target = torch.arange(num_labels, dtype=torch.float32, device=labels.device)
    return torch.softmax(
        -(target.unsqueeze(0) - labels.unsqueeze(1)).abs() * peakiness, dim=1
    )



from pathlib import Path
from typing import List
import string

from torch import Tensor
import numpy as np
import onnx
import onnxruntime
import pytest
import torch

from sihl.heads import AutoregressiveTextRecognition

ALPHABET = string.ascii_lowercase
BATCH_SIZE, NUM_CHANNELS, HEIGHT, WIDTH = 4, 256, 128, 128
MAX_SEQUENCE_LENGTH, LEVEL = 16, 5
ONNX_VERSION, ONNX_FILE_NAME = 18, "autoregressive_text_recognition.onnx"


@pytest.fixture()
def model() -> AutoregressiveTextRecognition:
    return AutoregressiveTextRecognition(
        in_channels=[3] + [NUM_CHANNELS for _ in range(LEVEL)],
        num_tokens=len(ALPHABET),
        max_sequence_length=MAX_SEQUENCE_LENGTH,
    )


@pytest.fixture()
def backbone_output() -> List[Tensor]:
    return [torch.randn((BATCH_SIZE, 3, HEIGHT, WIDTH))] + [
        torch.randn((BATCH_SIZE, NUM_CHANNELS, HEIGHT // 2**_, WIDTH // 2**_))
        for _ in range(LEVEL)
    ]


@pytest.fixture()
def targets() -> List[List[str]]:
    return [
        torch.tensor([ALPHABET.index(_) for _ in text])
        for text in "lorem ipsum dolor sit amet".split(" ")[:BATCH_SIZE]
    ]


def test_forward(
    model: AutoregressiveTextRecognition, backbone_output: List[Tensor]
) -> None:
    tokens = model(backbone_output)
    assert tuple(tokens.shape) == (BATCH_SIZE, model.max_sequence_length)


def test_training_step(
    model: AutoregressiveTextRecognition,
    backbone_output: List[Tensor],
    targets: List[List[str]],
) -> None:
    loss, _ = model.training_step(backbone_output, targets)
    assert loss.item()


def test_validation_step(
    model: AutoregressiveTextRecognition,
    backbone_output: List[Tensor],
    targets: List[List[str]],
) -> None:
    model.on_validation_start()
    loss, _ = model.validation_step(backbone_output, targets)
    assert loss.item()
    metrics = model.on_validation_end()
    assert metrics


@pytest.fixture()
def onnx_model(
    model: AutoregressiveTextRecognition, backbone_output: List[Tensor]
) -> None:
    torch.onnx.export(
        model,
        args=backbone_output,
        f=ONNX_FILE_NAME,
        opset_version=ONNX_VERSION,
        input_names=[f"input_level_{idx}" for idx in range(len(backbone_output))],
        output_names=[f"head0/{name}" for name, shape in model.output_shapes.items()],
        dynamic_axes=dict(
            {
                f"input_level_{lvl}": {
                    0: "batch_size",
                    2: f"height/{2**lvl}",
                    3: f"width/{2**lvl}",
                }
                for lvl in range(len(backbone_output))
            },
            **{
                f"head0/{name}": {
                    shape_idx: str(shape_value)
                    for shape_idx, shape_value in enumerate(shape)
                }
                for name, shape in model.output_shapes.items()
            },
        ),
        external_data=False,
        verify=True,
        # dynamo=True,
        # report=True,
    )
    onnx_model = onnx.load(ONNX_FILE_NAME)
    Path(ONNX_FILE_NAME).unlink()
    return onnx_model


def test_onnx_inference(
    onnx_model, model: AutoregressiveTextRecognition, backbone_output: List[Tensor]
) -> None:
    model.eval()
    onnx_session = onnxruntime.InferenceSession(onnx_model.SerializeToString())
    onnx_input = {
        f"input_level_{idx}": _.numpy()
        for idx, _ in enumerate(backbone_output)
        if f"input_level_{idx}" in [node.name for node in onnx_model.graph.input]
    }
    # just check that 99% of values are equal.
    pytorch_output = model(backbone_output).detach().numpy()
    onnx_output = onnx_session.run(None, onnx_input)[0]
    assert np.sum(np.abs(pytorch_output - onnx_output) > 1) / onnx_output.size < 0.01



from typing import List, Type

from math import log

from torch import Tensor, nn
import timm
import torch

from sihl.layers import AntialiasedDownscaler, Normalize
from sihl.utils import interpolate
from sihl.torchvision_backbone import freeze_levels

TIMM_BACKBONE_NAMES = (
    "convnext_atto",
    "convnext_base",
    "convnext_femto",
    "convnext_large",
    "convnext_nano",
    "convnext_pico",
    "convnext_small",
    "convnext_tiny",
    "convnext_xlarge",
    "convnext_xxlarge",
    "convnextv2_atto",
    "convnextv2_base",
    "convnextv2_femto",
    # "convnextv2_huge",
    "convnextv2_large",
    "convnextv2_nano",
    "convnextv2_pico",
    # "convnextv2_small",  # missing weights?!
    "convnextv2_tiny",
    "dla34",
    "dla60",
    "dla102",
    "dla169",
    "efficientnet_b0",
    "efficientnet_b1",
    "efficientnet_b2",
    "efficientnet_b3",
    "efficientnet_b4",
    "efficientnet_b5",
    # "efficientnet_b6",
    # "efficientnet_b7",
    # "efficientnet_b8",
    "efficientnet_lite0",
    # "efficientnet_lite1",
    # "efficientnet_lite2",
    # "efficientnet_lite3",
    # "efficientnet_lite4",
    # "efficientnetv2_l",
    # "efficientnetv2_m",
    # "efficientnetv2_s",
    # "efficientnetv2_xl",
    "hrnet_w18",
    "hrnet_w30",
    "hrnet_w32",
    "hrnet_w40",
    "hrnet_w44",
    "hrnet_w48",
    "hrnet_w64",
    # "mobilenetv2_035",
    "mobilenetv2_050",
    # "mobilenetv2_075",
    "mobilenetv2_100",
    "mobilenetv2_140",
    # "mobilenetv3_large_075",
    "mobilenetv3_large_100",
    "mobilenetv3_small_050",
    "mobilenetv3_small_075",
    "mobilenetv3_small_100",
    "mobilenetv4_conv_large",
    "mobilenetv4_conv_medium",
    "mobilenetv4_conv_small",
    "mobilenetv4_hybrid_large",
    # "mobilenetv4_hybrid_large_075",
    "mobilenetv4_hybrid_medium",
    # "mobilenetv4_hybrid_medium_075",
    "resnet18",
    "resnet26",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    # "resnet200",
    "resnetv2_50",
    "resnetv2_101",
    # "resnetv2_152",
)


class TimmBackbone(nn.Module):
    """https://github.com/huggingface/pytorch-image-models"""

    def __init__(
        self,
        name: str,
        pretrained: bool = False,
        input_channels: int = 3,
        top_level: int = 5,
        frozen_levels: int = 0,
        freeze_batchnorms: bool = False,
        downscaler: Type[nn.Module] = AntialiasedDownscaler,
    ) -> None:
        """
        Args:
            name (str): Architecture name (must be in TIMM_BACKBONE_NAMES)
            pretrained (bool, optional): Use weights from Imagenet pretraining. Defaults to False.
            input_channels (int, optional): Number of input image channels. Defaults to 3.
            top_level (int, optional): Deepest level (with stride 2^level). Defaults to 5.
            frozen_levels (int, optional): How many levels to freeze (from level 0). Defaults to 0.
            freeze_batchnorms (bool, optional): Whether to freeze batchnorm running stats. Defaults to False.
            downscaler (Type[nn.Module], optional): Downscaler module used for levels above 5. Defaults to AntialiasedDownscaler.
        """
        super().__init__()
        self.name = name
        self.top_level = top_level
        try:
            self.model = timm.create_model(
                self.name,
                features_only=True,
                scriptable=True,
                exportable=True,
                pretrained=pretrained,
                in_chans=input_channels,
            )
        except KeyError as error:
            raise ValueError(
                f"Architecture {name} is not supported. "
                f"Select from {TIMM_BACKBONE_NAMES}"
            ) from error
        except RuntimeError as error:
            raise ValueError("Failed to create TimmBackbone") from error
        reds = self.model.feature_info.reduction()
        assert all(log(reduction, 2).is_integer() for reduction in reds)

        self.normalize = nn.Identity()
        if pretrained and input_channels == 3:
            self.normalize = Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            )

        # Some models don't have a level 1 (stide 2), so we manually "add" it
        # FIXME: this is a hack
        if 2 not in reds:
            old_forward = self.model.forward

            def new_forward(self, x):
                size = (x.shape[2] // 2, x.shape[3] // 2)
                return [interpolate(x, size=size)] + old_forward(x)

            self.model.forward = new_forward.__get__(self.model)

        # freeze modules in first `frozen_levels` levels
        if pretrained:
            level_names = self.model.feature_info.module_name()
            # FIXME: this is a hack
            if "convnext" in name:
                level_names = [_.replace(".", "_") for _ in level_names]
            freeze_levels(self.model, frozen_levels, level_names, freeze_batchnorms)

        min_size = 2 ** (self.top_level + 1)
        self.dummy_input = torch.zeros(1, input_channels, min_size, min_size)
        self.out_channels = [input_channels] + [
            _.shape[1] for _ in self.model.forward(self.dummy_input)
        ]
        out_channels = self.out_channels[-1]
        extra_layers = range(self.top_level - 5)
        self.out_channels += [out_channels for _ in extra_layers]
        self.downscalers = nn.ModuleList(
            [AntialiasedDownscaler(out_channels, out_channels) for _ in extra_layers]
        )

    def forward(self, input: Tensor) -> List[Tensor]:
        assert input.shape[2] % 2**self.top_level == 0
        assert input.shape[3] % 2**self.top_level == 0
        x = self.normalize(input)
        outputs = self.model.forward(x)
        outputs = [input] + [
            interpolate(output, size=(x.shape[2] // 2**level, x.shape[3] // 2**level))
            for output, level in zip(outputs, range(1, self.top_level + 1))
        ]
        for downscaler in self.downscalers:
            outputs.append(downscaler(outputs[-1]))
        return outputs



from typing import List, Tuple, Dict

from einops import rearrange, reduce
from torch import nn, Tensor
from torch.nn import functional
from torchmetrics import MeanMetric
from torchmetrics.text import CharErrorRate, EditDistance
import torch

from sihl.layers import ConvNormAct, SequentialConvBlocks, SimpleUpscaler
from sihl.heads.semantic_segmentation import SPPM, UAFM


class SceneTextRecognition(nn.Module):
    """Scene text recognition is predicting a sequence of tokens. Tokens can represent
    textual characters, but not necessarily. The prediction is performed in parallel,
    which is fast, but can struggle to predict very long sequences.
    """

    def __init__(
        self,
        in_channels: List[int],
        num_tokens: int,
        max_sequence_length: int,
        bottom_level: int = 3,
        top_level: int = 5,
        num_channels: int = 256,
        num_layers: int = 3,
        pool_sizes: List[int] = [1, 2],
        logitnorm_scale: float = 20,
    ) -> None:
        """
        Args:
            in_channels (List[int]): Number of channels in input feature maps, sorted by level.
            num_tokens (int): Number of possible tokens.
            max_sequence_length (int): Maximum length of predicted sequences.
            level (int, optional): Level of inputs this head is attached to. Defaults to 3.
            num_channels (int, optional): Number of convolutional channels. Defaults to 256.
            num_layers (int, optional): Number of convolutional layers. Defaults to 4.
        """
        assert 0 < num_tokens < 256
        assert max_sequence_length > 0
        assert len(in_channels) > top_level >= bottom_level > 0
        assert num_channels > 0 and num_layers >= 0
        super().__init__()

        self.in_channels = in_channels
        self.num_tokens = num_tokens
        self.max_sequence_length = max_sequence_length
        self.pool_sizes = pool_sizes
        self.bottom_level = bottom_level
        self.top_level = top_level
        self.levels = range(bottom_level, top_level + 1)
        self.rev_levels = list(reversed(range(bottom_level, top_level)))
        self.logitnorm_scale = logitnorm_scale
        self.threshold = 0.5

        self.context_aggregation = SPPM(
            in_channels[top_level], num_channels, pool_sizes, with_shortcut=True
        )
        self.lateral_convs = nn.ModuleList(
            [ConvNormAct(in_channels[level], num_channels) for level in self.rev_levels]
        )
        self.upscalers = nn.ModuleList(
            [SimpleUpscaler(num_channels, num_channels) for level in self.rev_levels]
        )
        self.fusions = nn.ModuleList(
            [UAFM(num_channels, num_channels) for level in self.rev_levels]
        )

        self.stem = SequentialConvBlocks(num_channels, num_channels, num_layers)
        self.class_head = nn.Conv2d(num_channels, num_tokens + 1, 1)
        self.index_head = ConvNormAct(
            num_channels, max_sequence_length, 1, norm=None, act="sigmoid"
        )
        self.output_shapes = {
            "scores": ("batch_size", max_sequence_length),
            "tokens": ("batch_size", max_sequence_length),
        }

    def get_features(self, inputs: List[Tensor]) -> Tensor:
        x = self.context_aggregation(inputs[self.top_level])
        for level, lateral, upscale, fuse in zip(
            self.rev_levels, self.lateral_convs, self.upscalers, self.fusions
        ):
            x = fuse(lateral(inputs[level]), upscale(x))
        return self.stem(x)

    def get_maps(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor]:
        x = self.get_features(inputs)
        index_map = self.index_head(x)
        class_map = self.class_head(x).softmax(dim=1)
        return index_map, class_map

    def forward(self, inputs: List[Tensor]) -> Tuple[Tensor, Tensor]:
        x = self.get_features(inputs)
        index_map = rearrange(self.index_head(x), "n l h w -> n 1 l h w")
        class_map = rearrange(self.class_head(x), "n k h w -> n k 1 h w")
        logits = reduce(class_map * index_map, " n k l h w -> n k l", "mean")
        logits = functional.normalize(logits, dim=1) * self.logitnorm_scale
        scores = logits.softmax(dim=1)
        scores, pred_tokens = scores.max(dim=1)  # (N, L), (N, L)
        paddings = pred_tokens == self.num_tokens
        scores[paddings], pred_tokens[paddings] = 0, 0
        return scores, pred_tokens

    def training_step(
        self, inputs: List[Tensor], tokens: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        x = self.get_features(inputs)
        index_map = rearrange(self.index_head(x), "n l h w -> n 1 l h w")
        class_map = rearrange(self.class_head(x), "n k h w -> n k 1 h w")
        logits = reduce(class_map * index_map, " n k l h w -> n k l", "mean")
        logits = functional.normalize(logits, dim=1) * self.logitnorm_scale
        target_shape = (logits.shape[0], logits.shape[2])
        target = torch.full(target_shape, self.num_tokens, device=logits.device)
        for batch_idx, sample_tokens in enumerate(tokens):
            for char_pos, token_idx in enumerate(sample_tokens):
                target[batch_idx, char_pos] = token_idx
        loss = functional.cross_entropy(logits, target)
        return loss, {}

    def on_validation_start(self) -> None:
        self.token_error_rate = CharErrorRate()
        self.edit_distance = EditDistance()
        self.loss_computer = MeanMetric(nan_strategy="ignore")
        self.matches: List[bool] = []

    def validation_step(
        self, inputs: List[Tensor], tokens: List[Tensor]
    ) -> Tuple[Tensor, Dict[str, float]]:
        scores, pred_tokens = self.forward(inputs)
        # HACK: using miscellaneous symbols unicode block as 256 unique tokens
        predictions = [
            "".join(
                chr(0x2600 + token.item())
                for score, token in zip(sample_scores, sample_tokens)
                if score > self.threshold and token != self.num_tokens
            ).strip()
            for sample_scores, sample_tokens in zip(scores, pred_tokens)
        ]
        ground_truths = [
            "".join(chr(0x2600 + token.item()) for token in label).strip()
            for label in tokens
        ]
        self.token_error_rate.update(predictions, ground_truths)
        self.edit_distance.update(predictions, ground_truths)
        self.matches.extend(
            [pred == gt for pred, gt in zip(predictions, ground_truths)]
        )
        loss, metrics = self.training_step(inputs, tokens)
        self.loss_computer.to(loss.device).update(loss)
        return loss, metrics

    def on_validation_end(self) -> Dict[str, float]:
        return {
            "loss": self.loss_computer.compute().item(),
            "token_error_rate": self.token_error_rate.compute().item(),
            "edit_distance": self.edit_distance.compute().item(),
            "accuracy": sum(self.matches) / len(self.matches),
        }



from typing import List

from matplotlib import pyplot as plt
import numpy as np
import torch

from sihl.heads import ViewInvarianceLearning

from .common import get_images, plot_to_numpy


@get_images.register(ViewInvarianceLearning)
def _(head, config, input, target, features) -> List[np.ndarray]:
    cross_correlation = head.get_correlation(features, target)
    target = torch.eye(cross_correlation.shape[0], device=cross_correlation.device)
    cross_correlation = (1 - cross_correlation.abs()).to("cpu").numpy()
    target = (1 - target).to("cpu").numpy()
    image = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()[0]
    visualizations = []
    fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
    for ax in axes:
        ax.axis("off")
    axes[0].title.set_text("Input")
    axes[0].imshow(image)
    axes[1].title.set_text("Target")
    axes[1].imshow(target, cmap="gray")
    axes[2].title.set_text("Prediction")
    axes[2].imshow(cross_correlation, cmap="gray", vmin=0, vmax=1)
    fig.tight_layout()
    visualizations.append(plot_to_numpy(fig))
    plt.close()
    return visualizations



from typing import List

from matplotlib import patches
from matplotlib import pyplot as plt
from torchvision.ops import masks_to_boxes
from torchvision.transforms.v2 import Resize
import numpy as np
import torch

from sihl.heads import InstanceSegmentation
from sihl.utils import edges

from .common import get_images, plot_to_numpy, COLORS


@get_images.register(InstanceSegmentation)
def _(head, config, input, target, features) -> List[np.ndarray]:
    # # remove empty masks
    # for idx in range(len(target["masks"])):
    #     if target["masks"][idx].shape[0] != 0:
    #         keep = target["masks"][idx].sum(dim=(1, 2)).to(torch.bool)
    #         target["masks"][idx] = target["masks"][idx][keep]
    #         target["classes"][idx] = target["classes"][idx][keep]
    boxes = [masks_to_boxes(sample_masks) for sample_masks in target["masks"]]

    categories = config["categories"] if "categories" in config else None
    prediction = head(features)
    if prediction is not None:
        num_instances, pred_scores, pred_classes, pred_masks = prediction
    resize = Resize(target["masks"][0].shape[1:], antialias=False)
    images = (resize(input).permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch, image in enumerate(images):
        seen_categories = []
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.set_xticks([])
            ax.set_yticks([])
        axes[0].title.set_text("Input")
        axes[0].axis("off")
        axes[0].imshow(image)

        # sort masks by size -> draw large masks first
        axes[1].title.set_text("Target")
        if target is not None:
            for label, mask, box in zip(
                target["classes"][batch], target["masks"][batch], boxes[batch]
            ):
                category = str(label) if categories is None else categories[label]
                if category not in seen_categories:
                    seen_categories.append(category)
                color = COLORS[seen_categories.index(category) % len(COLORS)]
                axes[1].imshow(get_mask_img(mask, category, color))
                # axes[1].add_patch(get_patch(label.to("cpu"), box.to("cpu")))
        axes[2].title.set_text("Prediction")
        axes[2].imshow(np.full_like(image, fill_value=255))
        if prediction is not None:
            n = max(5, num_instances[batch])
            for label, mask in zip(pred_classes[batch, :n], pred_masks[batch, :n]):
                category = str(label) if categories is None else categories[label]
                legend = None
                if category not in seen_categories:
                    seen_categories.append(category)
                    legend = category
                color = COLORS[seen_categories.index(category) % len(COLORS)]
                axes[2].imshow(get_mask_img(mask, category, color))
                try:
                    axes[2].add_patch(
                        get_patch(
                            masks_to_boxes(mask.unsqueeze(0)).to("cpu"),
                            label.to("cpu"),
                            color,
                            legend,
                        )
                    )
                except Exception:
                    pass

        if len(seen_categories) > 0:
            fig.legend(
                handles=[
                    patches.Patch(
                        color=[_ / 255 for _ in COLORS[i % len(COLORS)] + (128,)],
                        label=label,
                    )
                    for i, label in enumerate(seen_categories)
                ],
                loc="lower center",
                frameon=False,
                ncol=min(7, len(seen_categories)),
            )
        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations


def get_mask_img(mask, category, color):
    mask_img = np.zeros((mask.shape[0], mask.shape[1], 4), dtype=np.uint8)
    mask = (mask > 0.5).to(torch.float).to("cpu")
    edge_mask = edges(mask[None, None, :, :])[0, 0] > 0.5
    mask[edge_mask] = 1
    mask_img[..., 0] = mask * color[0]
    mask_img[..., 1] = mask * color[1]
    mask_img[..., 2] = mask * color[2]
    mask_img[..., 3] = mask * 128
    mask_img[..., 3][edge_mask] = 255
    return mask_img


def get_patch(label, box, color, legend):
    return patches.Rectangle(
        (box[0], box[1]),
        box[2] - box[0],
        box[3] - box[1],
        linewidth=1,
        edgecolor=[_ / 255 for _ in color],
        facecolor="none",
        label=legend,
    )



from pathlib import Path
from typing import Tuple, List
import logging
import random

from lightning.pytorch.callbacks import Callback
from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision.io.image import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import MetricLearning


class QuerySetCallback(Callback):
    def __init__(self, index_set_dataloader: DataLoader) -> None:
        super().__init__()
        self.index_set_dataloader = index_set_dataloader

    def on_validation_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        assert isinstance(pl_module.model.heads[0], MetricLearning)
        with torch.no_grad():
            pl_module.model.heads[0].reset_validation_index_set()
            for batch in self.index_set_dataloader:
                inputs = pl_module.model.extract_features(batch[0].to(pl_module.device))
                pl_module.model.heads[0].extend_validation_index_set(
                    inputs, batch[1].to(pl_module.device)
                )

    def on_validation_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        pl_module.model.heads[0].reset_validation_index_set()


class StanfordCarsDataset(torch.utils.data.Dataset[Tuple[Tensor, Tensor]]):
    def __init__(self, data_dir: Path, ids: List[str], split: str) -> None:
        self.split = split
        self.data_dir = data_dir
        self.ids = ids
        self.images: List[Tuple[Path, int]] = []
        if split == "train":
            self.transform = transforms.Compose(
                [
                    transforms.Resize((256, 256)),
                    transforms.RandomHorizontalFlip(),
                    transforms.ColorJitter(
                        brightness=0.5, contrast=0.5, saturation=0.5, hue=0.5
                    ),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transform = transforms.Compose(
                [
                    transforms.Resize((256, 256)),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

        for img in data_dir.rglob("*.jpg"):
            annot = img.parent.name
            if annot in ids:
                self.images.append((img, self.ids.index(annot)))

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, target = self.images[idx]
        image = torchvision.io.read_image(
            str(image_path.resolve()), mode=ImageReadMode.RGB
        )
        return self.transform(image), torch.tensor(target)

    def __len__(self) -> int:
        return len(self.images)


class StanfordCarsDataModule(pl.LightningDataModule):
    def __init__(self, batch_size) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "StanfordCars"
        with open(self.data_dir / "names.csv") as f:
            ids = f.read().splitlines()
        ids = [_.replace("/", "-") for _ in ids]
        random.shuffle(ids)
        self.ids = ids

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "jutrera/stanford-car-dataset-by-classes-folder",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        split = len(self.ids) // 2
        self.trainset = StanfordCarsDataset(
            self.data_dir, self.ids[split:], split="train"
        )
        self.validset = StanfordCarsDataset(
            self.data_dir, self.ids[:split], split="valid"
        )

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            drop_last=True,
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.validset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            drop_last=True,
        )


if __name__ == "__main__":
    logging.basicConfig(
        level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="metric_learning"
    )
    lightning.seed_everything(0)

    datamodule = StanfordCarsDataModule(batch_size=128)
    datamodule.prepare_data()
    datamodule.setup()

    trainer = pl.Trainer(
        max_epochs=10,
        accelerator="gpu",
        logger=logger,
        callbacks=[
            pl.callbacks.RichProgressBar(leave=True),
            QuerySetCallback(datamodule.val_dataloader()),
        ],
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        head = MetricLearning(
            in_channels=backbone.out_channels,
            num_identities=len(datamodule.ids),
            embedding_dim=1024,
            margin=1 / len(datamodule.ids),
        )
        model = SihlLightningModule(
            model=SihlModel(backbone=backbone, neck=None, heads=[head]),
            optimizer=torch.optim.AdamW,
            optimizer_kwargs={"lr": 1e-4, "weight_decay": 1e-5},
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=datamodule)



from pathlib import Path
from typing import Tuple
import json
import logging

from faster_coco_eval.core.mask import decode as mask_decode
from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
from torchvision import tv_tensors
from torchvision.io import ImageReadMode
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.layers import FPN
from sihl.heads import SemanticSegmentation


class CocoSemanticSegmentationDataset(torch.utils.data.Dataset):
    CATEGORIES = [
        "background",
        "banner",
        "blanket",
        "branch",
        "bridge",
        "building-other",
        "bush",
        "cabinet",
        "cage",
        "cardboard",
        "carpet",
        "ceiling-other",
        "ceiling-tile",
        "cloth",
        "clothes",
        "clouds",
        "counter",
        "cupboard",
        "curtain",
        "desk-stuff",
        "dirt",
        "door-stuff",
        "fence",
        "floor-marble",
        "floor-other",
        "floor-stone",
        "floor-tile",
        "floor-wood",
        "flower",
        "fog",
        "food-other",
        "fruit",
        "furniture-other",
        "grass",
        "gravel",
        "ground-other",
        "hill",
        "house",
        "leaves",
        "light",
        "mat",
        "metal",
        "mirror-stuff",
        "moss",
        "mountain",
        "mud",
        "napkin",
        "net",
        "paper",
        "pavement",
        "pillow",
        "plant-other",
        "plastic",
        "platform",
        "playingfield",
        "railing",
        "railroad",
        "river",
        "road",
        "rock",
        "roof",
        "rug",
        "salad",
        "sand",
        "sea",
        "shelf",
        "sky-other",
        "skyscraper",
        "snow",
        "solid-other",
        "stairs",
        "stone",
        "straw",
        "structural-other",
        "table",
        "tent",
        "textile-other",
        "towel",
        "tree",
        "vegetable",
        "wall-brick",
        "wall-concrete",
        "wall-other",
        "wall-panel",
        "wall-stone",
        "wall-tile",
        "wall-wood",
        "water-other",
        "waterdrops",
        "window-blind",
        "window-other",
        "wood",
        "other",
    ]

    def __init__(self, data_dir: Path, train: bool = False) -> None:
        self.image_size = 640
        self.train = train
        self.data_dir = data_dir
        self.annots_by_image = {}
        split = "train" if train else "val"
        self.image_dir = self.data_dir / f"{split}2017"

        with open(
            self.data_dir / "stuff_annotations_trainval2017" / f"stuff_{split}2017.json"
        ) as f:
            coco_data = json.load(f)

        image_by_id = {
            image_annot["id"]: image_annot["file_name"]
            for image_annot in coco_data["images"]
        }
        for annot in coco_data["annotations"]:
            image_name = image_by_id[annot["image_id"]]
            image_path = str((self.image_dir / image_name).resolve())
            if image_path not in self.annots_by_image:
                self.annots_by_image[image_path] = []
            self.annots_by_image[image_path].append(annot)
        self.annots_by_image = [
            (key, value) for key, value in self.annots_by_image.items()
        ]

        if self.train:
            self.transform = transforms.Compose(
                [
                    transforms.ScaleJitter(
                        (self.image_size, self.image_size), scale_range=(0.5, 2.0)
                    ),
                    transforms.RandomCrop(self.image_size, pad_if_needed=True),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )
        else:
            self.transform = transforms.Compose(
                [
                    transforms.Resize(self.image_size - 1, max_size=self.image_size),
                    transforms.RandomCrop(self.image_size, pad_if_needed=True),
                    transforms.ToDtype(torch.float32, scale=True),
                ]
            )

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        image_path, annots = self.annots_by_image[idx]
        image = torchvision.io.read_image(image_path, mode=ImageReadMode.RGB)
        target = torch.stack(
            [
                torch.from_numpy(mask_decode(annot["segmentation"]))
                * (annot["category_id"] - 91)
                for annot in annots
            ]
        ).sum(dim=0)
        image, target = self.transform(image, tv_tensors.Mask(target))
        return image, target

    def __len__(self) -> int:
        return len(self.annots_by_image)


class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int = 32) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "coco_2017"
        self.data_dir.mkdir(exist_ok=True)

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "clkmuhammed/microsoft-coco-2017-common-objects-in-context",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        self.trainset = CocoSemanticSegmentationDataset(self.data_dir, train=True)
        self.validset = CocoSemanticSegmentationDataset(self.data_dir, train=False)

    def train_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset, batch_size=self.batch_size, shuffle=True, num_workers=4
        )

    def val_dataloader(self) -> DataLoader[Tuple[Tensor, Tensor]]:
        return DataLoader(
            self.validset, batch_size=self.batch_size, shuffle=False, num_workers=4
        )


if __name__ == "__main__":
    FORMAT = "%(message)s"
    logging.basicConfig(
        level="INFO", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="semantic_segmentation"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_steps=90_000,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=1,
        val_check_interval=0.2,
        precision="16-mixed",
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True)
        neck = FPN(backbone.out_channels, 256, bottom_level=3, top_level=5)
        head = SemanticSegmentation(
            in_channels=neck.out_channels,
            num_classes=len(CocoSemanticSegmentationDataset.CATEGORIES),
            ignore_index=CocoSemanticSegmentationDataset.CATEGORIES.index("other"),
            bottom_level=3,
            top_level=5,
            num_layers=0,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=torch.optim.AdamW,
            optimizer_kwargs={"lr": 1e-4, "weight_decay": 1e-4},
            data_config={
                "categories": CocoSemanticSegmentationDataset.CATEGORIES,
                "ignore": CocoSemanticSegmentationDataset.CATEGORIES.index("other"),
            },
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=16))
